import inspect
from typing import Any, Hashable, List, Optional

from .._internal import API


@API.public
class AntidoteError(Exception):
    """ Base class of all errors of antidote. """

    def __repr__(self):
        return f"{type(self).__name__}({self})"


@API.public
class DuplicateDependencyError(AntidoteError):
    """
    A dependency already exists with the same id.
    *May* be raised by _providers.
    """
    message: Optional[str]

    def __init__(self, dependency_or_message: Hashable, existing_definition: Any = None):
        if isinstance(dependency_or_message, str) and existing_definition is None:
            self.message = dependency_or_message
        else:
            self.message = None
            self.dependency = dependency_or_message
            self.existing_definition = existing_definition

    def __str__(self):
        if self.message is not None:
            return self.message
        else:
            return f"The dependency {self.dependency} already exists. " \
                   f"It points to {self.existing_definition}."


@API.public
class DependencyInstantiationError(AntidoteError):
    """
    The dependency could not be instantiated.
    """

    def __init__(self, dependency):
        super().__init__(f"Could not instantiate {dependency}")


@API.public
class DependencyCycleError(AntidoteError):
    """
    A dependency cycle is found.
    """

    def __init__(self, dependencies: List):
        self.dependencies = dependencies

    def __str__(self):
        return "Cycle:" + ''.join(map(self._repr, enumerate(self.dependencies))) + "\n"

    @staticmethod
    def _repr(i_dep) -> str:
        index, dependency = i_dep
        if inspect.isclass(dependency):
            return f"\n    {index}: {dependency.__module__}" \
                   f".{getattr(dependency, '__name__', dependency)}"

        return f"\n    {index}: {repr(dependency)}"


@API.public
class DependencyNotFoundError(AntidoteError):
    """
    The dependency could not be found.
    """

    def __init__(self, dependency: Hashable):
        self.missing_dependency = dependency

    def __str__(self):
        return repr(self.missing_dependency)


@API.public
class FrozenWorldError(AntidoteError):
    """
    An action failed because the world is frozen. Typically happens when trying
    to register a dependency after having called freeze() on the world.
    """


@API.private
class DebugNotAvailableError(AntidoteError):
    """
    Currently provider do not have to implement the debug behavior. If not, this error
    will be raised and discarded (a warning may be emitted).
    """



from . import world
from .constants import const, Constants
from .core import inject, wire, Wiring
from .factory import factory, Factory
from .implementation import Implementation, implementation
from .lazy import LazyCall, LazyMethodCall
from .service import service, Service
from .tag import Tag, Tagged
from .utils import is_compiled


def __version__() -> str:  # pragma: no cover
    import pkg_resources
    try:
        return pkg_resources.get_distribution(__name__).version
    except pkg_resources.DistributionNotFound:
        # package is not installed
        return ''


__all__ = ['world', 'const', 'Constants', 'inject', 'wire', 'Wiring', 'factory',
           'Factory', 'Implementation', 'implementation', 'LazyCall', 'LazyMethodCall',
           'service', 'Service', 'Tag', 'Tagged', 'is_compiled']



import inspect
from typing import Dict

from .slots import SlotsRepr
from .. import API
from ..._compatibility.typing import GenericMeta


@API.private
class ImmutableMeta(type):
    def __new__(mcls, name, bases, namespace, abstract: bool = False, **kwargs):
        if '__slots__' not in namespace:
            raise TypeError("Attributes must be defined in slots")
        slots = set(namespace['__slots__'])
        if any(name.startswith('__') for name in slots):
            raise ValueError("Private attributes are not supported.")

        if abstract:
            if len(slots) > 0:
                raise ValueError("Cannot be abstract and have a non-empty __slots__")

        # TODO: Type ignore necessary when type checking with Python 3.6
        #       To be removed ASAP.
        return super().__new__(mcls, name, bases, namespace, **kwargs)  # type: ignore


# TODO: remove after Python 3.6 support drops.
@API.private
class ImmutableGenericMeta(ImmutableMeta, GenericMeta):
    pass


@API.private
class Immutable(SlotsRepr, metaclass=ImmutableMeta, abstract=True):
    """
    Imitates immutable behavior by raising an exception when modifying an
    attribute through the standard means.
    """
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        # quick way to initialize an Immutable through args. It won't take into
        # account parent classes though.
        attrs: Dict[str, object] = dict(zip(self.__slots__, args))
        attrs.update(kwargs)
        for attr, value in attrs.items():
            object.__setattr__(self, attr, value)

    def __setattr__(self, name, value):
        raise AttributeError(f"{type(self)} is immutable")


@API.private
class FinalImmutableMeta(ImmutableMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        for b in bases:
            if isinstance(b, ImmutableMeta) \
                    and b.__module__ != __name__:
                raise TypeError(f"Type '{b.__name__}' cannot be inherited by {name}.")

        return super().__new__(mcls, name, bases, namespace, **kwargs)


@API.private
class FinalImmutable(Immutable, metaclass=FinalImmutableMeta, abstract=True):
    __slots__ = ()



import functools
from typing import Callable, Hashable, List, Sequence

from . import API
from .utils import FinalImmutable
from ..core.container import Container
from ..core.exceptions import DependencyNotFoundError

compiled = False


@API.private
class Injection(FinalImmutable):
    """
    Maps an argument name to its dependency and if the injection is required,
    which is equivalent to no default argument.
    """
    __slots__ = ('arg_name', 'required', 'dependency')
    arg_name: str
    required: bool
    dependency: Hashable


@API.private
class InjectionBlueprint(FinalImmutable):
    """
    Stores all the injections for a function.
    """
    __slots__ = ('injections',)
    injections: Sequence[Injection]

    def is_empty(self):
        return all(injection.dependency is None for injection in self.injections)


@API.private
def build_wrapper(blueprint: InjectionBlueprint,
                  wrapped: Callable,
                  skip_self: bool = False):
    """Used for consistency with Cython implementation."""
    return InjectedWrapper(blueprint, wrapped, skip_self)


@API.private
def get_wrapper_dependencies(wrapper: 'InjectedWrapper') -> List[Hashable]:
    if not isinstance(wrapper, InjectedWrapper):
        raise TypeError(f"Argument must be an {InjectedWrapper}")

    blueprint: InjectionBlueprint = getattr(wrapper,
                                            f"_{InjectedWrapper.__name__}__blueprint")
    return [inj.dependency for inj in blueprint.injections if inj.dependency is not None]


@API.private
def is_wrapper(x) -> bool:
    return isinstance(x, InjectedWrapper)


@API.private
class InjectedWrapper:
    """
    Wrapper which injects all the dependencies not supplied in the passed
    arguments. An InjectionBlueprint is used to store the mapping of the
    arguments to their dependency if any and if the injection is required.
    """

    def __init__(self,
                 blueprint: InjectionBlueprint,
                 wrapped: Callable,
                 skip_self: bool = False):
        """
        Args:
            blueprint: Injection blueprint for the underlying function
            wrapped:  real function to be called
            skip_self:  whether the first argument must be skipped. Used internally
        """
        self.__blueprint = blueprint
        self.__wrapped__ = wrapped
        self.__injection_offset = 1 if skip_self else 0
        functools.wraps(wrapped, updated=())(self)

    def __call__(self, *args, **kwargs):
        from .state import get_container
        kwargs = _inject_kwargs(
            get_container(),
            self.__blueprint,
            self.__injection_offset + len(args),
            kwargs
        )
        return self.__wrapped__(*args, **kwargs)

    def __get__(self, instance, owner):
        return InjectedBoundWrapper(
            self.__blueprint,
            self.__wrapped__.__get__(instance, owner),  # type: ignore
            isinstance(self.__wrapped__, classmethod)
            or (not isinstance(self.__wrapped__, staticmethod) and instance is not None)
        )

    def __getattr__(self, item):
        return getattr(self.__wrapped__, item)


@API.private
class InjectedBoundWrapper(InjectedWrapper):
    """
    Behaves like Python bound methods. Unsure whether this is really necessary
    or not.
    """

    def __get__(self, instance, owner):
        return self  # pragma: no cover


@API.private
def _inject_kwargs(container: Container,
                   blueprint: InjectionBlueprint,
                   offset: int,
                   kwargs: dict) -> dict:
    """
    Does the actual injection of the dependencies. Used by InjectedCallableWrapper.
    """
    dirty_kwargs = False
    for injection in blueprint.injections[offset:]:
        if injection.dependency is not None and injection.arg_name not in kwargs:
            try:
                arg = container.get(injection.dependency)
                if not dirty_kwargs:
                    kwargs = kwargs.copy()
                    dirty_kwargs = True
                kwargs[injection.arg_name] = arg
            except DependencyNotFoundError:
                if injection.required:
                    raise

    return kwargs



import pytest

from antidote import world
from antidote._providers.lazy import FastLazyConst, Lazy, LazyProvider
from antidote.core import Container, DependencyInstance


@pytest.fixture()
def lazy_provider():
    with world.test.empty():
        world.provider(LazyProvider)
        yield world.get(LazyProvider)


class Dummy(Lazy):
    def __init__(self, value, singleton=False):
        self.value = value
        self.singleton = singleton

    def lazy_get(self, container: Container) -> DependencyInstance:
        return DependencyInstance(container.get(self.value), singleton=self.singleton)


def test_lazy():
    with world.test.empty():
        x = object()
        world.singletons.add(x, object())
        lazy_provider = LazyProvider()

        assert world.test.maybe_provide_from(lazy_provider,
                                             Dummy(x)).value is world.get(x)
        for s in [True, False]:
            assert world.test.maybe_provide_from(lazy_provider,
                                                 Dummy(x, singleton=s)).singleton is s


def test_exists():
    lazy = LazyProvider()
    assert not lazy.exists(object())
    assert lazy.exists(Dummy('x'))


def test_fast_lazy_method(lazy_provider: LazyProvider):
    class A:
        def method(self, value):
            return self, value

    world.singletons.add(A, A())
    x = object()

    (owner, value) = world.get(FastLazyConst('X', A, 'method', x, lambda v: v))
    assert owner is world.get(A)
    assert value is x

    assert world.get(FastLazyConst('X', A, 'method', x, lambda v: 1)) == 1


def test_copy(lazy_provider: LazyProvider):
    assert isinstance(lazy_provider.clone(True), LazyProvider)
    assert isinstance(lazy_provider.clone(False), LazyProvider)



import textwrap

from antidote import (Constants, factory, Implementation, implementation, inject,
                      LazyCall, LazyMethodCall, Service, Tag, world)
from antidote._internal.utils import raw_getattr, short_id


class DebugTestCase:
    def __init__(self, value, expected: str, depth: int = -1):
        if expected.startswith('\n'):
            expected = textwrap.dedent(expected[1:])
        lines = expected.splitlines(keepends=True)
        while not lines[-1].strip():
            lines.pop()

        self.value = value
        self.expected = ''.join(lines)
        self.depth = depth


def assert_valid(*test_cases: DebugTestCase):
    for test_case in test_cases:
        assert world.debug(test_case.value,
                           depth=test_case.depth) == test_case.expected


def test_no_debug():
    def g():
        pass

    assert_valid(
        DebugTestCase(
            value=g,
            expected=f"""{g!r} is neither a dependency nor is anything injected."""
        )
    )


def test_implementation_debug():
    class Interface:
        pass

    class Dummy(Interface, Service):
        pass

    prefix = "tests.world.test_debug.test_implementation_debug.<locals>"

    with world.test.clone():
        @implementation(Interface)
        def f():
            return Dummy

        assert_valid(DebugTestCase(
            value=Interface,
            expected=f"""
                Permanent link: {prefix}.Interface -> ??? defined by {prefix}.f
                    """
        ))

        world.get(Interface)
        assert_valid(DebugTestCase(
            value=Interface,
            expected=f"""
                Permanent link: {prefix}.Interface -> {prefix}.Dummy defined by {prefix}.f
                └── {prefix}.Dummy
                    """
        ))

    with world.test.clone():
        undefined_expectations = f"""
            * Dynamic link: {prefix}.Interface -> ??? defined by {prefix}.g

            * = not singleton
        """

        @implementation(Interface, permanent=False)
        def g():
            return Dummy

        assert_valid(DebugTestCase(value=Interface,
                                   expected=undefined_expectations))

        world.get(Interface)
        assert_valid(DebugTestCase(value=Interface,
                                   expected=undefined_expectations))


def test_lazy_call_debug():
    prefix = "tests.world.test_debug.test_lazy_call_debug.<locals>"

    with world.test.new():
        def f():
            pass

        l1 = LazyCall(f)("arg", hello="world")
        assert_valid(
            DebugTestCase(
                value=l1,
                expected=f"""
                    Lazy: {prefix}.f(*('arg',), **{{'hello': 'world'}})  #{short_id(l1)}
                    """
            ),
            DebugTestCase(
                value=LazyCall(f, singleton=False)("arg", hello="world"),
                expected=f"""
                    * Lazy: {prefix}.f(*('arg',), **{{'hello': 'world'}})

                    * = not singleton
                    """
            ))

    with world.test.new():
        class MyService(Service):
            pass

        @inject
        def f(service: MyService):
            pass

        l2 = LazyCall(f)
        assert_valid(
            DebugTestCase(
                value=l2,
                expected=f"""
                    Lazy: {prefix}.f()  #{short_id(l2)}
                    └── {prefix}.f
                        └── {prefix}.MyService
                    """
            ),
            DebugTestCase(
                value=LazyCall(f, singleton=False),
                expected=f"""
                    * Lazy: {prefix}.f()
                    └── {prefix}.f
                        └── {prefix}.MyService

                    * = not singleton
                    """
            )
        )


def test_unknown_debug():
    prefix = "tests.world.test_debug.test_unknown_debug.<locals>"

    with world.test.new():
        class Service1:
            pass

        class Dummy(Service):
            def __init__(self, s: Service1):
                pass

        assert_valid(DebugTestCase(
            value=Dummy,
            expected=f"""
                {prefix}.Dummy
                └── /!\\ Unknown: {prefix}.Service1
            """
        ))


def test_singleton_debug():
    with world.test.new():
        world.singletons.add("test", 1)

        assert_valid(DebugTestCase(
            value="test",
            expected="""
                Singleton 'test' -> 1
            """
        ))


def test_wiring_debug():
    prefix = "tests.world.test_debug.test_wiring_debug.<locals>"

    with world.test.new():
        class Service1(Service):
            pass

        class Dummy(Service):
            __antidote__ = Service.Conf().with_wiring(methods=['__init__', 'get'])

            def __init__(self, s: Service1):
                pass

            def get(self, s: Service1):
                pass

        assert_valid(DebugTestCase(
            value=Dummy,
            expected=f"""
                {prefix}.Dummy
                ├── {prefix}.Service1
                └── Method: get
                    └── {prefix}.Service1
            """
        ))


def test_multiline_debug():
    prefix = "tests.world.test_debug.test_multiline_debug.<locals>"

    with world.test.new():
        class MultilineService(Service):
            @classmethod
            def __antidote_debug_repr__(cls):
                return "Multiline\nService"

        class Dummy(Service):
            def __init__(self, s: MultilineService):
                pass

        assert_valid(DebugTestCase(
            value=Dummy,
            expected=f"""
                {prefix}.Dummy
                └── Multiline
                    Service
            """
        ))


def test_tag():
    prefix = "tests.world.test_debug.test_tag.<locals>"

    with world.test.new():
        tag = Tag()

        class CustomTag(Tag):
            def group(self):
                return 'dummy'

        class S1(Service):
            __antidote__ = Service.Conf(tags=[tag, CustomTag()])

        assert_valid(
            DebugTestCase(
                value=tag,
                expected=f"""
                * Tag: Tag#{short_id(tag)}
                └── {prefix}.S1

                * = not singleton
                """
            ),
            DebugTestCase(
                value=CustomTag(),
                expected=f"""
                * Tag: 'dummy'
                └── {prefix}.S1

                * = not singleton
                """
            )
        )


def test_lazy_method_debug():
    prefix = "tests.world.test_debug.test_lazy_method_debug.<locals>"

    with world.test.new():
        class MyService(Service):
            pass

        class Conf(Service):
            @inject
            def fetch(self, value, service: MyService):
                return value

            DATA = LazyMethodCall(fetch)
            KW = LazyMethodCall(fetch)(value='1')
            DATA2 = LazyMethodCall(fetch, singleton=False)
            KW2 = LazyMethodCall(fetch, singleton=False)(value='2')

        assert_valid(
            DebugTestCase(
                value=Conf.DATA,
                expected=f"""
                    Lazy Method: fetch()  #{short_id(raw_getattr(Conf, 'DATA'))}
                    ├── {prefix}.Conf.fetch
                    │   └── {prefix}.MyService
                    └── {prefix}.Conf
                    """
            ),
            DebugTestCase(
                value=Conf.KW,
                expected=f"""
        Lazy Method: fetch(*(), **{{'value': '1'}})  #{short_id(raw_getattr(Conf, 'KW'))}
        ├── {prefix}.Conf.fetch
        │   └── {prefix}.MyService
        └── {prefix}.Conf
                    """
            ),
            DebugTestCase(
                value=Conf.DATA2,
                expected=f"""
                    * Lazy Method: fetch()
                    ├── {prefix}.Conf.fetch
                    │   └── {prefix}.MyService
                    └── {prefix}.Conf

                    * = not singleton
                    """
            ),
            DebugTestCase(
                value=Conf.KW2,
                expected=f"""
                    * Lazy Method: fetch(*(), **{{'value': '2'}})
                    ├── {prefix}.Conf.fetch
                    │   └── {prefix}.MyService
                    └── {prefix}.Conf

                    * = not singleton
                    """
            )
        )


def test_constants_debug():
    prefix = "tests.world.test_debug.test_constants_debug.<locals>"

    with world.test.new():
        class Conf(Constants):
            TEST = '1'

            def get(self, value):
                return value

        assert_valid(
            DebugTestCase(
                value=Conf.TEST,
                expected=f"""
            Const: {prefix}.Conf.TEST
            └── Lazy: {prefix}.Conf()  #{short_id(raw_getattr(Conf, 'TEST').dependency)}
                    """
            ))

    with world.test.new():
        class Conf(Constants):
            __antidote__ = Constants.Conf(public=True)

            TEST = '1'

            def get(self, value):
                return value

        assert_valid(
            DebugTestCase(
                value=Conf.TEST,
                expected=f"""
                    Const: {prefix}.Conf.TEST
                    └── {prefix}.Conf
                    """
            ))

    with world.test.new():
        class MyService(Service):
            pass

        class Conf(Constants):
            TEST = '1'

            @inject
            def get(self, value, service: MyService):
                return value

        assert_valid(
            DebugTestCase(
                value=Conf.TEST,
                expected=f"""
            Const: {prefix}.Conf.TEST
            ├── {prefix}.Conf.get
            │   └── {prefix}.MyService
            └── Lazy: {prefix}.Conf()  #{short_id(raw_getattr(Conf, 'TEST').dependency)}
                    """
            ))


def test_complex_debug():
    with world.test.new():
        tag = Tag()

        class Interface:
            pass

        class Service1(Service):
            pass

        class Service2:
            def __init__(self, service1: Service1):
                self.service1 = service1

        @factory
        def build_s2(service1: Service1) -> Service2:
            return Service2(service1)

        class Service3(Service):
            __antidote__ = Service.Conf(singleton=False,
                                        tags=[tag]).with_wiring(
                methods=['__init__', 'get'],
                dependencies=dict(
                    service2=Service2 @ build_s2))

            def __init__(self, service1: Service1, service2: Service2, i: Interface):
                self.service1 = service1
                self.service2 = service2
                self.i = i

            def get(self, service1: Service1):
                pass

            X = LazyMethodCall(get)

        class Service4(Interface, Implementation):
            __antidote__ = Implementation.Conf(tags=[tag]).with_wiring(
                dependencies=dict(service2=Service2 @ build_s2))

            def __init__(self, service1: Service1, service2: Service2,
                         service3: Service3):
                self.service1 = service1
                self.service2 = service2
                self.service3 = service3

        @inject
        def f(s: Service4):
            pass

        @inject(dependencies=[Service1.with_kwargs(test=1),
                              Service2 @ build_s2.with_kwargs(option=2)])
        def f_with_options(a, b):
            pass

        prefix = "tests.world.test_debug.test_complex_debug.<locals>"
        assert_valid(
            DebugTestCase(
                value=tag,
                depth=0,
                expected=f"""
                    * Tag: Tag#{short_id(tag)}

                    * = not singleton
                        """
            ),
            DebugTestCase(
                value=tag,
                depth=1,
                expected=f"""
                    * Tag: Tag#{short_id(tag)}
                    ├── {prefix}.Service4
                    └── * {prefix}.Service3

                    * = not singleton
                        """
            ),
            DebugTestCase(
                value=tag,
                depth=2,
                expected=f"""
                    * Tag: Tag#{short_id(tag)}
                    ├── {prefix}.Service4
                    │   ├── * {prefix}.Service3
                    │   ├── {prefix}.Service2 @ {prefix}.build_s2
                    │   └── {prefix}.Service1
                    └── * {prefix}.Service3
                        ├── Static link: {prefix}.Interface -> {prefix}.Service4
                        ├── {prefix}.Service2 @ {prefix}.build_s2
                        └── {prefix}.Service1

                    * = not singleton
                        """
            ),
            DebugTestCase(
                value=tag,
                depth=3,
                expected=f"""
                    * Tag: Tag#{short_id(tag)}
                    ├── {prefix}.Service4
                    │   ├── * {prefix}.Service3
                    │   │   ├── Static link: {prefix}.Interface -> {prefix}.Service4
                    │   │   ├── {prefix}.Service2 @ {prefix}.build_s2
                    │   │   └── {prefix}.Service1
                    │   ├── {prefix}.Service2 @ {prefix}.build_s2
                    │   │   └── {prefix}.build_s2
                    │   │       └── {prefix}.Service1
                    │   └── {prefix}.Service1
                    └── * {prefix}.Service3
                        ├── Static link: {prefix}.Interface -> {prefix}.Service4
                        │   └── {prefix}.Service4
                        ├── {prefix}.Service2 @ {prefix}.build_s2
                        │   └── {prefix}.build_s2
                        │       └── {prefix}.Service1
                        └── {prefix}.Service1

                    * = not singleton
                        """
            ),
            DebugTestCase(
                value=tag,
                expected=f"""
                    * Tag: Tag#{short_id(tag)}
                    ├── {prefix}.Service4
                    │   ├── * {prefix}.Service3
                    │   │   ├── Static link: {prefix}.Interface -> {prefix}.Service4
                    │   │   │   └── /!\\ Cyclic dependency: {prefix}.Service4
                    │   │   ├── {prefix}.Service2 @ {prefix}.build_s2
                    │   │   │   └── {prefix}.build_s2
                    │   │   │       └── {prefix}.Service1
                    │   │   └── {prefix}.Service1
                    │   ├── {prefix}.Service2 @ {prefix}.build_s2
                    │   │   └── {prefix}.build_s2
                    │   │       └── {prefix}.Service1
                    │   └── {prefix}.Service1
                    └── * {prefix}.Service3
                        ├── Static link: {prefix}.Interface -> {prefix}.Service4
                        │   └── {prefix}.Service4
                        │       ├── /!\\ Cyclic dependency: {prefix}.Service3
                        │       ├── {prefix}.Service2 @ {prefix}.build_s2
                        │       │   └── {prefix}.build_s2
                        │       │       └── {prefix}.Service1
                        │       └── {prefix}.Service1
                        ├── {prefix}.Service2 @ {prefix}.build_s2
                        │   └── {prefix}.build_s2
                        │       └── {prefix}.Service1
                        └── {prefix}.Service1

                    * = not singleton
                        """
            ),
            DebugTestCase(
                value=f,
                expected=f"""
                    {prefix}.f
                    └── {prefix}.Service4
                        ├── * {prefix}.Service3
                        │   ├── Static link: {prefix}.Interface -> {prefix}.Service4
                        │   │   └── /!\\ Cyclic dependency: {prefix}.Service4
                        │   ├── {prefix}.Service2 @ {prefix}.build_s2
                        │   │   └── {prefix}.build_s2
                        │   │       └── {prefix}.Service1
                        │   └── {prefix}.Service1
                        ├── {prefix}.Service2 @ {prefix}.build_s2
                        │   └── {prefix}.build_s2
                        │       └── {prefix}.Service1
                        └── {prefix}.Service1

                    * = not singleton
                        """
            ),
            DebugTestCase(
                value=f_with_options,
                expected=f"""
                    {prefix}.f_with_options
                    ├── {prefix}.Service2 @ {prefix}.build_s2(**{{'option': 2}})
                    │   └── {prefix}.build_s2
                    │       └── {prefix}.Service1
                    └── {prefix}.Service1(**{{'test': 1}})
                    """
            )
        )



import sys

if sys.version_info < (3, 8):
    from typing_extensions import final, Protocol
else:
    from typing import final, Protocol

if sys.version_info < (3, 7):
    from typing import GenericMeta
else:
    class GenericMeta(type):
        pass

__all__ = ['final', 'Protocol', 'GenericMeta']



from typing import Any, Callable, Hashable, Optional

import pytest

from antidote import world
from antidote.core import DependencyInstance
from antidote.core.exceptions import DuplicateDependencyError
from antidote.exceptions import DependencyNotFoundError
from ..core.utils import DummyProvider

sentinel = object()


@pytest.mark.parametrize('override', [
    pytest.param(lambda d, v: world.test.override.singleton(d, v), id='singleton'),
    pytest.param(lambda d, v: world.test.override.singleton({d: v}), id='singletons')
])
def test_singleton(override: Callable[[Any, Any], Any]):
    with world.test.empty():
        with world.test.clone(overridable=True):
            override('test', 'a')
            assert world.get("test") == 'a'

            # Still works, but has no impact as it has been overridden
            world.singletons.add('test', sentinel)
            assert world.get("test") == 'a'

        world.singletons.add("test", sentinel)

        with world.test.clone(overridable=True):
            override("test", 'a')
            assert world.get("test") == 'a'

            # double override works
            override("test", 'b')
            assert world.get("test") == 'b'

        assert world.get('test') is sentinel

        with world.test.clone(overridable=True, keep_singletons=True):
            assert world.get('test') is sentinel

            override("test", 'a')
            assert world.get("test") == 'a'

            with pytest.raises(DuplicateDependencyError):
                # already existed in the container without overrides.
                world.singletons.add("test", 2)

        assert world.get('test') is sentinel


def test_factory():
    # overrides normal provider
    with world.test.clone(overridable=True):
        world.provider(DummyProvider)
        p = world.get[DummyProvider]()
        p.data = {'test': sentinel}

        assert world.get('test') is sentinel

        @world.test.override.factory('test')
        def f():
            return 'a'

        assert world.get('test') == 'a'

        @world.test.override.factory('test')
        def g():
            return 'b'

        assert world.get('test') == 'b'

    # overrides singletons
    with world.test.clone(overridable=True):
        world.singletons.add("test", sentinel)
        assert world.get('test') is sentinel

        @world.test.override.factory('test')
        def f1():
            return 'a'

        assert world.get('test') == 'a'

    # non singleton
    with world.test.clone(overridable=True):
        @world.test.override.factory('test')
        def f2():
            return object()

        assert world.get('test') is not world.get('test')

    # singleton
    with world.test.clone(overridable=True):
        @world.test.override.factory('test', singleton=True)
        def f3():
            return object()

        assert world.get('test') is world.get('test')


def test_provider():
    # overrides normal provider
    with world.test.clone(overridable=True):
        world.provider(DummyProvider)
        p = world.get[DummyProvider]()
        p.data = {'test': sentinel, 'test2': sentinel}

        assert world.get('test') is sentinel
        assert world.get('test2') is sentinel

        @world.test.override.provider
        def f(dependency: Hashable) -> Optional[DependencyInstance]:
            if dependency == 'test':
                return DependencyInstance('a')

        assert world.get('test') == 'a'
        assert world.get('test2') is sentinel

    # overrides singletons
    with world.test.clone(overridable=True):
        world.singletons.add({'test': sentinel, 'test2': sentinel})

        assert world.get('test') is sentinel
        assert world.get('test2') is sentinel

        @world.test.override.provider
        def f2(dependency: Hashable) -> Optional[DependencyInstance]:
            if dependency == 'test':
                return DependencyInstance('a')

        assert world.get('test') == 'a'
        assert world.get('test2') is sentinel

    # non singleton dependency
    with world.test.clone(overridable=True):

        @world.test.override.provider
        def f3(dependency: Hashable) -> Optional[DependencyInstance]:
            if dependency == 'test':
                return DependencyInstance(object(), singleton=False)

        assert world.get('test') is not world.get('test')

    # singleton dependency
    with world.test.clone(overridable=True):

        @world.test.override.provider
        def f4(dependency: Hashable) -> Optional[DependencyInstance]:
            if dependency == 'test':
                return DependencyInstance(object(), singleton=True)

        assert world.get('test') is world.get('test')


def test_deep_clone():
    with world.test.empty():
        world.singletons.add("test", sentinel)

        with world.test.clone(overridable=True, keep_singletons=True):
            with world.test.clone(overridable=True, keep_singletons=False):
                with pytest.raises(DependencyNotFoundError):
                    world.get("test")

        with world.test.clone(overridable=True):
            world.test.override.singleton("test", 'a')
            with world.test.clone(overridable=True, keep_singletons=False):
                with pytest.raises(DependencyNotFoundError):
                    world.get("test")

            with world.test.clone(overridable=False, keep_singletons=False):
                with pytest.raises(DependencyNotFoundError):
                    world.get("test")

        with world.test.clone(overridable=True):
            world.test.override.singleton("test", 'a')
            with world.test.clone(overridable=True, keep_singletons=True):
                assert world.get("test") == 'a'

            with world.test.clone(overridable=False, keep_singletons=True):
                assert world.get("test") == 'a'


def test_debug():
    with world.test.clone(overridable=True):
        from antidote._internal import state

        world.singletons.add("test", sentinel)

        world.test.override.singleton("over", 1)

        @world.test.override.factory('factory')
        def f():
            return 2

        c = state.get_container()
        assert "Override" not in c.debug("test").info
        assert "Singleton" in c.debug("test").info

        assert "Override" in c.debug("over").info
        assert "Singleton" in c.debug("over").info

        assert "Override" in c.debug("factory").info
        assert "Factory" in c.debug("factory").info


def test_invalid():
    with world.test.clone(overridable=True):
        with pytest.raises(TypeError):
            world.test.override.singleton(1)

        with pytest.raises(TypeError):
            @world.test.override.factory('test', singleton=object())
            def f():
                pass

        with pytest.raises(TypeError):
            world.test.override.factory('test', singleton=True)(object())

        with pytest.raises(TypeError):
            world.test.override.provider(object())



from contextlib import contextmanager
from typing import Hashable, Optional

import pytest

from antidote import world
from antidote.core import (Container, DependencyInstance, does_not_freeze, Provider,
                           StatelessProvider)
from antidote.exceptions import FrozenWorldError


@contextmanager
def does_not_raise():
    yield


def test_freeze_world():
    class DummyProvider(Provider):
        def provide(self, dependency: Hashable, container: Container
                    ) -> Optional[DependencyInstance]:
            return None

        def clone(self, keep_singletons_cache: bool) -> 'DummyProvider':
            return self

        def register(self):
            return "register"

        @does_not_freeze
        def method(self):
            return "method"

        @staticmethod
        def static():
            return "static"

        @classmethod
        def klass(cls):
            return "klass"

    provider = DummyProvider()
    assert provider.register() == "register"
    assert provider.method() == "method"
    assert provider.static() == "static"
    assert provider.klass() == "klass"

    with world.test.empty():
        world.provider(DummyProvider)
        provider = world.get[DummyProvider]()
        assert provider.register() == "register"
        assert provider.method() == "method"
        assert provider.static() == "static"
        assert provider.klass() == "klass"

        world.freeze()
        assert provider.method() == "method"
        assert provider.static() == "static"
        assert provider.klass() == "klass"
        provider.clone(False)
        provider.provide(None, None)
        with pytest.raises(FrozenWorldError):
            provider.register()


def test_stateless():
    class DummyProvider(StatelessProvider):
        def provide(self, dependency: Hashable, container: Container
                    ) -> Optional[DependencyInstance]:
            return None

    p = DummyProvider()
    assert p.clone(True) is not p
    assert isinstance(p.clone(False), DummyProvider)


def test_no_default_implementation():
    class Dummy(Provider):
        pass

    with pytest.raises(NotImplementedError):
        Dummy().maybe_provide(object(), object())

    with pytest.raises(NotImplementedError):
        Dummy().exists(object())

    with pytest.raises(NotImplementedError):
        Dummy().clone(False)

    with pytest.raises(RuntimeError):
        Dummy().provide(object(), object())

    with pytest.raises(NotImplementedError):
        Dummy().exists(False)


def test_debug():
    x = object()

    class Dummy(Provider):
        def exists(self, dependency: Hashable) -> bool:
            return dependency is x

    Dummy().maybe_debug(object())  # should no fail
    with pytest.warns(UserWarning, match="(?i).*debug.*"):
        Dummy().maybe_debug(x)



from typing import Any

from ._internal import API
from ._internal.utils import AbstractMeta
from ._providers import ServiceProvider, TagProvider
from ._providers.service import Build
from .core import inject

_ABSTRACT_FLAG = '__antidote_abstract'


@API.private
class ServiceMeta(AbstractMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        cls = super().__new__(mcls, name, bases, namespace, **kwargs)
        if not kwargs.get('abstract'):
            _configure_service(cls)
        return cls

    @API.public
    def with_kwargs(cls, **kwargs) -> Any:
        """
        Creates a new dependency based on the service which will have the keyword
        arguments provided. If the service is a singleton and identical kwargs are used,
        the same instance will be given by Antidote.

        Args:
            **kwargs: Arguments passed on to :code:`__init__()`.

        Returns:
            Dependency to be retrieved from Antidote.
        """
        return Build(cls, kwargs)


@API.private
@inject
def _configure_service(cls,
                       service_provider: ServiceProvider,
                       tag_provider: TagProvider = None,
                       conf=None):
    from .service import Service

    conf = conf or getattr(cls, '__antidote__', None)
    if not isinstance(conf, Service.Conf):
        raise TypeError(f"Service configuration (__antidote__) is expected to be a "
                        f"{Service.Conf}, not a {type(conf)}")

    wiring = conf.wiring

    if wiring is not None:
        wiring.wire(cls)

    service_provider.register(cls, singleton=conf.singleton)

    if conf.tags is not None:
        if tag_provider is None:
            raise RuntimeError("No TagProvider registered, cannot use tags.")
        tag_provider.register(dependency=cls, tags=conf.tags)



import collections.abc as c_abc
import inspect
from typing import (Any, Callable, get_type_hints, Iterable, Optional,
                    overload, Tuple, TypeVar, Union)

from ._compatibility.typing import final, Protocol
from ._factory import FactoryMeta, LambdaFactory, PreBuild
from ._internal import API
from ._internal.utils import Copy, FinalImmutable
from ._providers import FactoryProvider, Tag, TagProvider
from .core.injection import DEPENDENCIES_TYPE, inject
from .core.wiring import Wiring, WithWiringMixin

F = TypeVar('F', bound=Callable[..., Any])


@API.private
class FactoryProtocol(Protocol[F]):
    """
    :meta private:
    """

    def __rmatmul__(self, dependency) -> Any:
        pass  # pragma: no cover

    def with_kwargs(self, **kwargs) -> PreBuild:
        pass  # pragma: no cover

    __call__: F


@API.public
class Factory(metaclass=FactoryMeta, abstract=True):
    """
    Abstract base class for a factory which provides a single dependency. The provided
    dependency is defined through the type annotation of :py:meth:`.__call__` which will
    be used to create the dependency.

    The factory instance will only be instantiated once, whether the dependency is a
    singleton, the default, or not. Both :py:meth:`.__init__` and :py:meth:`.__call__`
    are wired by default.

    To retrieve the dependency from Antidote you need to use a specific syntax
    :code:`dependency @ factory` as presented in the following examples. The goal of it is
    twofold:

    - Ensure that the factory is loaded whenever you require the dependency.
    - Better maintainability as you know *where* the dependency comes from.

    .. note::

        If you only need a simple function, consider using :py:func:`.factory` instead.

    .. doctest:: helpers_Factory

        >>> from antidote import Factory, world
        >>> class ExternalService:
        ...     pass
        >>> class MyFactory(Factory):
        ...     def __call__(self) -> ExternalService:
        ...         return ExternalService()
        >>> world.get[ExternalService @ MyFactory]()
        <ExternalService ...>

    For customization use :py:attr:`.__antidote__`:

    .. doctest:: helpers_Factory_v2

        >>> from antidote import Factory, world
        >>> class ExternalService:
        ...     pass
        >>> class MyFactory(Factory):
        ...     __antidote__ = Factory.Conf(singleton=False) \\
        ...         .with_wiring(use_names=True)
        ...
        ...     def __call__(self) -> ExternalService:
        ...         return ExternalService()

    One can customize the instantiation and use the same service with different
    configuration:

    .. doctest:: helpers_Factory_v3

        >>> from antidote import Factory, world, inject
        >>> class ExternalService:
        ...     def __init__(self, name):
        ...         self.name = name
        >>> class MyFactory(Factory):
        ...     def __call__(self, name = 'default') -> ExternalService:
        ...         return ExternalService(name)
        >>> world.get[ExternalService](ExternalService @ MyFactory).name
        'default'
        >>> s = world.get[ExternalService](
        ...     ExternalService @ MyFactory.with_kwargs(name='perfection'))
        >>> s.name
        'perfection'
        >>> # The same instance will be returned for those keywords as MyFactory was
        ... # declared as returning a singleton.
        ... s is world.get(ExternalService @ MyFactory.with_kwargs(name='perfection'))
        True
        >>> # You can also keep the dependency and re-use it
        ... PerfectionService = ExternalService @ MyFactory.with_kwargs(name='perfection')
        >>> @inject(dependencies=dict(service=PerfectionService))
        ... def f(service):
        ...     return service
        >>> f() is s
        True

    """

    @final
    class Conf(FinalImmutable, WithWiringMixin):
        """
        Immutable factory configuration. To change parameters on a existing instance, use
        either method :py:meth:`.copy` or
        :py:meth:`.core.wiring.WithWiringMixin.with_wiring`.
        """
        __slots__ = ('wiring', 'singleton', 'tags', 'public')
        wiring: Optional[Wiring]
        singleton: bool
        tags: Optional[Tuple[Tag]]
        public: bool

        def __init__(self,
                     *,
                     wiring: Optional[Wiring] = Wiring(
                         methods=['__call__', '__init__'],
                         ignore_missing_method=['__init__']
                     ),
                     singleton: bool = True,
                     tags: Optional[Iterable[Tag]] = None,
                     public: bool = False):
            """

            Args:
                wiring: Wiring to be applied on the factory. By default only
                    :code:`__init__()` and :code:`__call__()` will be wired. To deactivate
                    any wiring at all use :py:obj:`None`.
                singleton: Whether the returned dependency is a singleton or not. If yes,
                    the factory will only be called once and its result cached. Defaults
                    to :py:obj:`True`.
                tags: Iterable of :py:class:`~.._providers.tag.Tag` tagging to the
                      provided dependency.
                public: Whether the factory instance should be retrievable through
                    Antidote or not. Defaults to :py:obj:`False`
            """
            if not isinstance(public, bool):
                raise TypeError(f"public must be a boolean, "
                                f"but not a {type(public)}")
            if not isinstance(singleton, bool):
                raise TypeError(f"singleton must be a boolean, "
                                f"but not a {type(singleton)}")
            if not (tags is None or isinstance(tags, c_abc.Iterable)):
                raise TypeError(f"tags must be None or an iterable of strings/Tags, "
                                f"but not a {type(tags)}")
            elif tags is not None:
                tags = tuple(tags)
                if not all(isinstance(t, Tag) for t in tags):
                    raise TypeError(f"Not all tags were instances of Tag: {tags}")
            if not (wiring is None or isinstance(wiring, Wiring)):
                raise TypeError(f"wiring must be None or a Wiring, "
                                f"but not a {type(wiring)}")
            super().__init__(wiring=wiring, singleton=singleton, tags=tags, public=public)

        def copy(self,
                 *,
                 wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL,
                 singleton: Union[bool, Copy] = Copy.IDENTICAL,
                 tags: Union[Optional[Iterable[Tag]], Copy] = Copy.IDENTICAL,
                 public: Union[bool, Copy] = Copy.IDENTICAL):
            """
            Copies current configuration and overrides only specified arguments.
            Accepts the same arguments as :py:meth:`.__init__`
            """
            return Factory.Conf(
                wiring=self.wiring if wiring is Copy.IDENTICAL else wiring,
                singleton=self.singleton if singleton is Copy.IDENTICAL else singleton,
                tags=self.tags if tags is Copy.IDENTICAL else tags,
                public=self.public if public is Copy.IDENTICAL else public,
            )

    __antidote__: Conf = Conf()
    """
    Configuration of the factory. Defaults to wire :py:meth:`.__init__` and
    :py:meth:`.__call__`.
    """


@overload
def factory(f: F,  # noqa: E704  # pragma: no cover
            *,
            auto_wire: bool = True,
            singleton: bool = True,
            dependencies: DEPENDENCIES_TYPE = None,
            use_names: Union[bool, Iterable[str]] = None,
            use_type_hints: Union[bool, Iterable[str]] = None,
            tags: Iterable[Tag] = None
            ) -> FactoryProtocol[F]: ...


@overload
def factory(*,  # noqa: E704  # pragma: no cover
            auto_wire: bool = True,
            singleton: bool = True,
            dependencies: DEPENDENCIES_TYPE = None,
            use_names: Union[bool, Iterable[str]] = None,
            use_type_hints: Union[bool, Iterable[str]] = None,
            tags: Iterable[Tag] = None
            ) -> Callable[[F], FactoryProtocol[F]]: ...


@API.public
def factory(f: F = None,
            *,
            auto_wire: bool = True,
            singleton: bool = True,
            dependencies: DEPENDENCIES_TYPE = None,
            use_names: Union[bool, Iterable[str]] = None,
            use_type_hints: Union[bool, Iterable[str]] = None,
            tags: Iterable[Tag] = None
            ):
    """
    Registers a factory which provides as single dependency, defined through the return
    type annotation.

    To retrieve the dependency from Antidote you need to use a specific syntax
    :code:`dependency @ factory` as presented in the following examples. The goal of it is
    twofold:

    - Ensure that the factory is loaded whenever you require the dependency.
    - Better maintainability as you know *where* the dependency comes from.

    .. note::

        If you need a stateful factory or want to implement a complex one prefer using
        :py:class:`.Factory` instead.

    .. doctest:: helpers_factory

        >>> from antidote import factory, world
        >>> class ExternalService:
        ...     pass
        >>> @factory
        ... def build_service() -> ExternalService:
        ...     return ExternalService()
        >>> world.get[ExternalService](ExternalService @ build_service)
        <ExternalService ...>

    One can customize the instantiation and use the same service with different
    configuration:

    .. doctest:: helpers_factory_v2

        >>> from antidote import factory, world, inject
        >>> class ExternalService:
        ...     def __init__(self, name):
        ...         self.name = name
        >>> @factory
        ... def build_service(name = 'default') -> ExternalService:
        ...     return ExternalService(name)
        >>> world.get[ExternalService](ExternalService @ build_service).name
        'default'
        >>> s = world.get[ExternalService](
        ...     ExternalService @ build_service.with_kwargs(name='perfection'))
        >>> s.name
        'perfection'
        >>> # The same instance will be returned for those keywords as MyFactory was
        ... # declared as returning a singleton.
        ... s is world.get(ExternalService @ build_service.with_kwargs(name='perfection'))
        True
        >>> # You can also keep the dependency and re-use it
        ... PerfectionService = \\
        ...     ExternalService @ build_service.with_kwargs(name='perfection')
        >>> @inject(dependencies=dict(service=PerfectionService))
        ... def f(service):
        ...     return service
        >>> f() is s
        True

    Args:
        f: Callable which builds the dependency.
        singleton: If True, `func` will only be called once. If not it is
            called at each injection.
        auto_wire: Whether the function should have its arguments injected or not
            with :py:func:`~.injection.inject`.
        dependencies: Propagated to :py:func:`~.injection.inject`.
        use_names: Propagated to :py:func:`~.injection.inject`.
        use_type_hints: Propagated to :py:func:`~.injection.inject`.
        tags: Iterable of :py:class:`~.._providers.tag.Tag` applied to the provided
            dependency.

    Returns:
        The factory or the function decorator.

    """
    if not (auto_wire is None or isinstance(auto_wire, bool)):
        raise TypeError(f"auto_wire can be None or a boolean, not {type(auto_wire)}")

    @inject
    def register_factory(func,
                         factory_provider: FactoryProvider,
                         tag_provider: TagProvider = None):
        if not inspect.isfunction(func):
            raise TypeError(f"{func} is not a function")

        dependency = get_type_hints(func).get('return')
        if dependency is None:
            raise ValueError("A return annotation is necessary. "
                             "It is used a the dependency.")
        if not inspect.isclass(dependency):
            raise TypeError(f"The return annotation is expected to be a class, "
                            f"not {type(dependency)}.")

        if auto_wire:
            func = inject(func,
                          dependencies=dependencies,
                          use_names=use_names,
                          use_type_hints=use_type_hints)

        factory_id = factory_provider.register(factory=func,
                                               singleton=singleton,
                                               dependency=dependency)

        if tags is not None:
            if tag_provider is None:
                raise RuntimeError("No TagProvider registered, cannot use tags.")
            tag_provider.register(dependency=factory_id, tags=tags)

        return LambdaFactory(func, factory_id)

    return f and register_factory(f) or register_factory



import itertools
from typing import Any, Callable, Optional, Tuple, Type

import pytest

from antidote import Constants, Factory, Implementation, Service, world
from antidote._compatibility.typing import Protocol
from antidote.core import Wiring


class Interface:
    pass


class FactoryOutput:
    pass


class FactoryOutput2:
    pass


class A:
    pass


class B:
    pass


@pytest.fixture(autouse=True)
def new_world():
    with world.test.new():
        world.singletons.add({A: A(),
                              B: B(),
                              'a': object(),
                              'b': object(),
                              'x': object(),
                              'y': object(),
                              'z': object()})
        yield


class DummyProtocol(Protocol):
    a: A
    b: B

    def __init__(self, a: A = None, b: B = None):
        pass

    def method(self, a: A = None, b: B = None) -> Tuple[A, B]:
        pass

    def method2(self, a=None, b=None) -> Tuple[Any, Any]:
        pass

    def method3(self, x=None, y=None, z=None) -> Tuple[Any, Any, Any]:
        pass


default_wiring = object()


def builder(cls: type, wiring_kind: str, subclass: bool = False):
    meta_kwargs = dict(abstract=True) if subclass else dict()
    if cls is Implementation:
        bases = (Interface, Implementation)
    else:
        bases = (cls,)

    def build(wiring: Wiring = None):
        class Dummy(*bases, **meta_kwargs):
            if wiring is not default_wiring:
                if wiring is not None:
                    if wiring_kind == 'Wiring':
                        __antidote__ = cls.Conf(wiring=wiring)
                    else:
                        __antidote__ = cls.Conf().with_wiring(**{
                            attr: getattr(wiring, attr)
                            for attr in Wiring.__slots__
                        })
                else:
                    __antidote__ = cls.Conf(wiring=None)

            def __init__(self, a: A = None, b: B = None):
                super().__init__()
                self.a = a
                self.b = b

            def method(self, a: A = None, b: B = None) -> Tuple[A, B]:
                return a, b

            def method2(self, a=None, b=None) -> Tuple[Any, Any]:
                return a, b

            def method3(self, x=None, y=None, z=None) -> Tuple[Any, Any, Any]:
                return x, y, z

            def __call__(self) -> FactoryOutput:  # for Factory
                pass

            def get(self):  # for Constants
                pass

        if subclass:
            class SubDummy(Dummy):
                def __call__(self) -> FactoryOutput2:  # for Factory
                    pass

            return SubDummy
        else:
            return Dummy

    return build


@pytest.fixture(params=[
    pytest.param((c, w), id=f"{c.__name__} - w")
    for (c, w) in itertools.product([Factory,
                                     Service,
                                     Constants,
                                     Implementation],
                                    ['with_wiring', 'Wiring'])
])
def class_builder(request):
    (cls, wiring_kind) = request.param
    return builder(cls, wiring_kind)


@pytest.fixture(params=[
    pytest.param((c, w), id=f"{c.__name__} - w")
    for (c, w) in itertools.product([Factory,
                                     Service,
                                     Constants],
                                    ['with_wiring', 'Wiring'])
])
def subclass_builder(request):
    (cls, wiring_kind) = request.param
    return builder(cls, wiring_kind, subclass=True)


F = Callable[[Optional[Wiring]], Type[DummyProtocol]]


def test_auto_wiring(class_builder: F):
    """__init__ should be injected by default."""
    dummy = class_builder(default_wiring)()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)


def test_no_wiring(class_builder: F):
    """__init__ should be injected by default."""
    dummy = class_builder(None)()
    assert dummy.a is None
    assert dummy.b is None


def test_method_wiring(class_builder: F):
    dummy = class_builder(Wiring(methods=('method',)))()
    assert dummy.a is None
    assert dummy.b is None

    (a, b) = dummy.method()
    assert a is world.get(A)
    assert b is world.get(B)


def test_ignore_missing_method(class_builder: F):
    with pytest.raises(AttributeError, match=".*unknown.*"):
        class_builder(Wiring(methods=('unknown',)))()

    with pytest.raises(AttributeError, match=".*unknown.*"):
        class_builder(Wiring(methods=('unknown',), ignore_missing_method=False))()

    with pytest.raises(AttributeError, match=".*unknown.*"):
        class_builder(Wiring(methods=['unknown', 'method'],
                             ignore_missing_method=['method']))()

    with world.test.clone(keep_singletons=True):
        class_builder(Wiring(methods=('unknown',),
                             ignore_missing_method=['unknown']))()

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('unknown',), ignore_missing_method=True))()
        assert dummy.a is None
        assert dummy.b is None

        (a, b) = dummy.method()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',),
                                     ignore_missing_method=['method']))()
        assert dummy.a is None
        assert dummy.b is None

        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)


def test_super_wiring(subclass_builder: F):
    with pytest.raises(AttributeError, match=".*method.*"):
        # method() is not implemented by the subclass.
        subclass_builder(Wiring(methods=('method',)))()

    with world.test.clone(keep_singletons=True):
        sub_dummy = subclass_builder(Wiring(methods=('method',),
                                            ignore_missing_method=True))()
        assert sub_dummy.a is None
        assert sub_dummy.b is None

        (a, b) = sub_dummy.method()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        sub_dummy = subclass_builder(Wiring(methods=('method',),
                                            ignore_missing_method=True,
                                            wire_super=False))()
        assert sub_dummy.a is None
        assert sub_dummy.b is None

        (a, b) = sub_dummy.method()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        sub_dummy = subclass_builder(Wiring(methods=('method', '__init__'),
                                            wire_super=True))()
        assert sub_dummy.a is world.get(A)
        assert sub_dummy.b is world.get(B)

        (a, b) = sub_dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

    with world.test.clone(keep_singletons=True):
        sub_dummy = subclass_builder(Wiring(methods=('method', '__init__'),
                                            wire_super=('method', '__init__')))()
        assert sub_dummy.a is world.get(A)
        assert sub_dummy.b is world.get(B)

        (a, b) = sub_dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

    with world.test.clone(keep_singletons=True):
        sub_dummy = subclass_builder(Wiring(methods=('method', '__init__'),
                                            wire_super=('__init__',),
                                            ignore_missing_method=True))()
        assert sub_dummy.a is world.get(A)
        assert sub_dummy.b is world.get(B)

        (a, b) = sub_dummy.method()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        sub_dummy = subclass_builder(Wiring(methods=('method', '__init__'),
                                            wire_super=('method',),
                                            ignore_missing_method=True))()
        assert sub_dummy.a is None
        assert sub_dummy.b is None

        (a, b) = sub_dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

    with pytest.raises(ValueError, match=".*methods.*"):
        subclass_builder(Wiring(methods=('method', '__init__'),
                                wire_super=('__init__', 'unknown')))()


def test_use_names(class_builder: F):
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method2',)))()
        (a, b) = dummy.method2()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method2',), use_names=False))()
        (a, b) = dummy.method2()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method2',), use_names=True))()
        (a, b) = dummy.method2()
        assert a is world.get('a')
        assert b is world.get('b')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method2',), use_names=('a',)))()
        (a, b) = dummy.method2()
        assert a is world.get('a')
        assert b is None

    # Does not override type_hints
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',), use_names=True))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)


def test_use_type_hints(class_builder: F):
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',)))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',), use_type_hints=True))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',), use_type_hints=False))()
        (a, b) = dummy.method()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',), use_type_hints=('a',)))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is None

    # with use_names
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',),
                                     use_type_hints=False,
                                     use_names=True))()
        (a, b) = dummy.method()
        assert a is world.get('a')
        assert b is world.get('b')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method',),
                                     use_type_hints=('a',),
                                     use_names=True))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get('b')


def test_dependencies_dict(class_builder: F):
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     dependencies=dict()))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

        (a, b) = dummy.method2()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     use_names=True,
                                     dependencies=dict(a='x', b='y')))()
        (a, b) = dummy.method()
        assert a is world.get('x')
        assert b is world.get('y')

        (a, b) = dummy.method2()
        assert a is world.get('x')
        assert b is world.get('y')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     use_names=True,
                                     dependencies=dict(b='y')))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get('y')

        (a, b) = dummy.method2()
        assert a is world.get('a')
        assert b is world.get('y')


def test_dependencies_seq(class_builder: F):
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     dependencies=tuple()))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

        (a, b) = dummy.method2()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     dependencies=[None, None]))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

        (a, b) = dummy.method2()
        assert a is None
        assert b is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     use_names=True,
                                     dependencies=['x', 'y']))()
        (a, b) = dummy.method()
        assert a is world.get('x')
        assert b is world.get('y')

        (a, b) = dummy.method2()
        assert a is world.get('x')
        assert b is world.get('y')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     use_names=True,
                                     dependencies=[None, 'y']))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get('y')

        (a, b) = dummy.method2()
        assert a is world.get('a')
        assert b is world.get('y')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     use_names=True,
                                     dependencies=['x', None]))()
        (a, b) = dummy.method()
        assert a is world.get('x')
        assert b is world.get(B)

        (a, b) = dummy.method2()
        assert a is world.get('x')
        assert b is world.get('b')


def test_dependencies_callable(class_builder: F):
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(
            methods=('method', 'method2'),
            dependencies=lambda arg: 'x' if arg.name == 'a' else 'y'))()
        (a, b) = dummy.method()
        assert a is world.get('x')
        assert b is world.get('y')

        (a, b) = dummy.method2()
        assert a is world.get('x')
        assert b is world.get('y')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(
            methods=('method', 'method2'),
            use_names=True,
            dependencies=lambda arg: 'x' if arg.name == 'a' else None))()
        (a, b) = dummy.method()
        assert a is world.get('x')
        assert b is world.get(B)

        (a, b) = dummy.method2()
        assert a is world.get('x')
        assert b is world.get('b')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     use_names=True,
                                     dependencies=lambda arg: None))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

        (a, b) = dummy.method2()
        assert a is world.get('a')
        assert b is world.get('b')


def test_dependencies_str(class_builder: F):
    with world.test.clone(keep_singletons=True):
        world.singletons.add({
            'conf:a': object(),
            'conf:b': object()
        })
        dummy = class_builder(Wiring(methods=('method', 'method2'),
                                     dependencies='conf:{arg_name}'))()
        (a, b) = dummy.method()
        assert a is world.get('conf:a')
        assert b is world.get('conf:b')

        (a, b) = dummy.method2()
        assert a is world.get('conf:a')
        assert b is world.get('conf:b')


def test_distinct_arguments(class_builder: F):
    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method3'),
                                     dependencies=['x', None, 'z']))()
        (a, b) = dummy.method()
        assert a is world.get('x')
        assert b is world.get(B)

        (x, y, z) = dummy.method3()
        assert x is world.get('x')
        assert y is None
        assert z is world.get('z')

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method3'),
                                     dependencies=dict(b='b', y='y')))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get('b')

        (x, y, z) = dummy.method3()
        assert x is None
        assert y is world.get('y')
        assert z is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method3'),
                                     use_type_hints=['a']))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is None

        (x, y, z) = dummy.method3()
        assert x is None
        assert y is None
        assert z is None

    with world.test.clone(keep_singletons=True):
        dummy = class_builder(Wiring(methods=('method', 'method3'),
                                     use_names=['x']))()
        (a, b) = dummy.method()
        assert a is world.get(A)
        assert b is world.get(B)

        (x, y, z) = dummy.method3()
        assert x is world.get('x')
        assert y is None
        assert z is None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Antidote documentation build configuration file, created by
# sphinx-quickstart on Sat Oct 14 22:07:20 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# sys.path.insert(0, os.path.abspath('..'))
# sys.path.insert(0, os.path.abspath('_themes'))

import antidote


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.intersphinx'
]

# Python code that is treated like it were put in a testcleanup directive for
# every file that is tested, and for every group.
doctest_global_setup = """
from antidote._internal import state
state.init()
"""
doctest_global_cleanup = """
from antidote._internal import state
state.reset()
"""

autodoc_member_order = 'bysource'
autoclass_content = "both"

# This config value contains the locations and names of other projects
# that should be linked to in this documentation.
intersphinx_mapping = {
    # 'python': ('https://docs.python.org/3.6', None),
    # 'attrs': ('http://www.attrs.org/en/stable', None)
}

# Prefix each section label with the name of the document it is in.
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Antidote'
copyright = '2017, Benjamin Rabier'
author = 'Benjamin Rabier'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tag.
release = antidote.__version__()
# The short X.Y version.
version = release.rsplit(".", 1)[0]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'flask_theme_support.FlaskyStyle'

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
add_module_names = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = "sphinx_rtd_theme"

# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {
#     'show_powered_by': False,
#     'github_user': 'Finistere',
#     'github_repo': 'antidote',
#     'github_banner': True,
#     'show_related': False,
#     'note_bg': '#FFF59C'
# }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
        'donate.html',
    ]
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Antidotedoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Antidote.tex', 'Antidote Documentation',
     'Benjamin Rabier', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'dependency manager', 'Antidote Documentation',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Antidote', 'Antidote Documentation',
     author, 'Antidote', 'One line description of project.',
     'Miscellaneous'),
]


# def do_not_skip_antidote(app, what, name, obj, skip, options):
#     return (name != "__antidote__") and skip


def setup(app):
    # app.connect("autodoc-skip-member", do_not_skip_antidote)
    app.add_css_file('css/style.css')  # may also be an URL



"""
Utilities for your Antidotes tests and your own.
"""
from . import override
from ._methods import clone, empty, maybe_provide_from, new

__all__ = ['override', 'clone', 'new', 'empty', 'maybe_provide_from']



def test_readme_simple() -> None:
    from antidote import inject, Service, Constants

    class Conf(Constants):
        # Configuration values are identified by public uppercase class
        # attributes. It helps refactoring as it's easy to find their usage
        # and their definition.
        # The Constants super class will treat their associated value as the input
        # argument of get(). This allows you to load lazily any configuration.
        DB_HOST = 'host'

        def __init__(self):
            self._data = {'host': 'localhost:6789'}

        def get(self, key):
            return self._data[key]

    # Declare Database as a dependency that can be injected
    class Database(Service):
        # Antidotes configuration for this class. `__init__()` is wired (injected) by
        # default.
        __antidote__ = Service.Conf().with_wiring(
            # Specifying that arguments named 'host' should be injected with the
            # dependency Conf.DB_HOST.
            dependencies=dict(host=Conf.DB_HOST))

        # You could have wired yourself `__init__()` by applying:
        # @inject(dependencies=dict(host=Conf.DB_HOST))
        def __init__(self, host: str):
            self._host = host

    # Inject dependencies in f(), by default only type annotations are used. But
    # arguments name, explicit mapping, etc.. can also be used.
    @inject
    def f(db: Database):
        pass

    f()  # Service will be automatically injected if not provided
    f(Database('localhost:6789'))  # but you can still use the function normally


def test_readme() -> None:
    """
    Simple example where a MovieDB interface is defined which can be used
    to retrieve the best movies. In our case the implementation uses IMDB
    to dot it.
    """
    from antidote import Constants, factory, Implementation, inject, world

    class MovieDB:
        """ Interface """

        def get_best_movies(self):
            pass

    class ImdbAPI:
        """ Class from an external library. """

        def __init__(self, *args, **kwargs):
            pass

    # Defining a singleton. Can only be overridden in tests.
    world.singletons.add('conf_path', '/...')

    class Conf(Constants):
        IMDB_HOST = 'imdb.host'
        IMDB_API_KEY = 'imdb.api_key'

        # `use_names=True` specifies that Antidote can use the argument names
        # when type hints are not present or too generic (builtins typically).
        __antidote__ = Constants.Conf().with_wiring(use_names=True)

        def __init__(self, conf_path: str):
            """ Load configuration from `conf_path` """
            self._raw_conf = {
                'imdb': {
                    'host': 'dummy_host',
                    'api_key': 'dummy_api_key'
                }
            }

        def get(self, key):
            """
            self.get('a.b') <=> self._raw_conf['a']['b']
            """
            from functools import reduce
            return reduce(dict.get, key.split('.'), self._raw_conf)  # type: ignore

    # ImdbAPI will be provided by this factory, as defined by the return type annotation.
    # The dependencies arguments specifies what must be injected
    @factory(dependencies=(Conf.IMDB_HOST, Conf.IMDB_API_KEY))
    def imdb_factory(host: str, api_key: str) -> ImdbAPI:
        # Here host = Conf().get('imdb.host')
        return ImdbAPI(host=host, api_key=api_key)

    # Implementation tells Antidote that this class should be used as an implementation of
    # the interface MovieDB
    class IMDBMovieDB(MovieDB, Implementation):
        # As ImdbAPI is provided by imdb_factory, Antidote requires it to be explicitly
        # specified. This ensures that can always track back where dependencies are
        # coming from.
        __antidote__ = Implementation.Conf().with_wiring(
            dependencies=dict(imdb_api=ImdbAPI @ imdb_factory))

        def __init__(self, imdb_api: ImdbAPI):
            self._imdb_api = imdb_api

        def get_best_movies(self):
            pass

    @inject
    def f(movie_db: MovieDB):
        pass

    # You can also retrieve dependencies by hand
    world.get[str](Conf.IMDB_HOST)  # the result will be cast to `str`
    # To avoid repetition, if the type is the dependency itself you can do:
    world.get[IMDBMovieDB]()

    # If you need to handle multiple different api_keys for some reason you can
    # specify them in the dependency itself:
    world.get[ImdbAPI](ImdbAPI @ imdb_factory.with_kwargs(api_key='XX'))
    # As imdb_factory returns a singleton, by default, this will also be the case
    # here. Using the same API key, will return the same instance. This avoids boilerplate
    # code when the same instance is needed with different arguments. The same works
    # with a Service. In the previous example you could have
    # used `Database.with_kwargs(host='something')`

    # Like before you can call f() without any arguments:
    f()

    # You can still explicitly pass the arguments to override
    # injection.
    conf = Conf('/path')
    f(IMDBMovieDB(imdb_factory(
        # The class attributes will retrieve the actual value when called on a instance.
        # Hence this is equivalent to conf.get('imdb.host'), making your tests easier.
        host=conf.IMDB_HOST,
        api_key=conf.IMDB_API_KEY,  # <=> conf.get('imdb.api_key')
    )))

    # When testing you can also override locally some dependencies:
    with world.test.clone(overridable=True, keep_singletons=True):
        world.test.override.singleton({
            Conf.IMDB_HOST: 'other host'
        })
        f()

    # If you encounter issues you can ask Antidote for a summary of what's happening
    # for a specific dependency. It becomes useful as an cycle/instantiation error
    # deep within the dependency tree results in a complex error stack.
    world.debug(f)
    """
    f
    └── Static link: MovieDB -> IMDBMovieDB
        └── IMDBMovieDB
            └── ImdbAPI @ imdb_factory
                └── imdb_factory
                    ├── Const: Conf.IMDB_API_KEY
                    │   └── Lazy: Conf()  #0BjHAQ
                    │       └── Singleton 'conf_path' -> '/...'
                    └── Const: Conf.IMDB_HOST
                        └── Lazy: Conf()  #0BjHAQ
                            └── Singleton 'conf_path' -> '/...'
    """

    # For example suppose we don't have the singleton `'conf_path'`
    with world.test.clone(keep_singletons=False):
        world.debug(f)
        # will output the following. As you can see, 'conf_path` is not found. Hence
        # when Conf will be instantiated it will fail.
        """
        f
        └── Static link: MovieDB -> IMDBMovieDB
            └── IMDBMovieDB
                └── ImdbAPI @ imdb_factory
                    └── imdb_factory
                        ├── Const: Conf.IMDB_API_KEY
                        │   └── Lazy: Conf()  #0BjHAQ
                        │       └── /!\\ Unknown: 'conf_path'
                        └── Const: Conf.IMDB_HOST
                            └── Lazy: Conf()  #0BjHAQ
                                └── /!\\ Unknown: 'conf_path'
        """



from .container import Container, DependencyInstance
from .injection import inject, validate_injection
from .provider import does_not_freeze, Provider, StatelessProvider
from .utils import Dependency
from .wiring import wire, Wiring



from contextlib import contextmanager
from typing import Any, Callable, Hashable, Optional, overload, TypeVar, Union

from ..._internal import API, state
from ...core.container import (DependencyInstance, OverridableRawContainer, RawContainer,
                               RawProvider)


@API.public
@contextmanager
def clone(*, keep_singletons: bool = False, overridable: bool = False):
    """
    Clone current Antidote state (singletons & providers) into a new container. It should
    be used when you need to rely on existing dependencies defined in your source code.
    Otherwise consider using :py:func:`.new`.

    You're safe to:
     - add / delete singletons
     - override dependencies if :code:`overridable=True`
     - register new dependencies

    .. doctest:: world_test_clone

        >>> from antidote import world, Service
        >>> from antidote.exceptions import DependencyNotFoundError
        >>> world.singletons.add("test", 1)
        >>> class MyService(Service):
        ...     pass
        >>> with world.test.clone():
        ...     # By default singletons are not propagated
        ...     world.get[int]("test")
        Traceback (most recent call last):
        ...
        DependencyNotFoundError: 'test'
        >>> with world.test.clone():
        ...     # But anything else will be accessible
        ...     world.get[MyService]()
        <MyService object ...>
        >>> # You can have the singletons if wish though:
        ... with world.test.clone(keep_singletons=True):
        ...     assert world.get[int]("test") == 1
        ...     # Overriding is still impossible by default.
        ...     world.singletons.add("test", 10)
        Traceback (most recent call last):
        ...
        DuplicateDependencyError: 'test'
        >>> with world.test.clone(overridable=True, keep_singletons=True):
        ...     # Antidote will support an override once before any usage.
        ...     world.test.override.singleton("test", 10)
        ...     world.get[int]("test")
        10
        >>> # The original singleton is preserved
        ... world.get[int]("test")
        1

    Args:
        keep_singletons: Whether current singletons should be kept in the new world. Keep
            in mind that while you can add and delete singletons without worries, keeping
            singletins means that you will have the same instances. Defaults to False.
        overridable: Whether dependencies can be overriden in the new container. Note
            that you can only override dependencies *once* and only *before any usage*.
            Defaults to False.

    Returns:
        context manager within you can make your tests.

    """
    if overridable:
        with state.override(lambda c: OverridableRawContainer.build(c, keep_singletons)):
            yield
    else:
        with state.override(lambda c: c.clone(keep_singletons=keep_singletons)):
            yield


@API.public
@contextmanager
def new(*, default_providers: bool = False):
    """
    Creates a new container void of any dependencies.

    Args:
        default_providers: Whether the default providers should be used. If not
            a new container with the same providers as before is created. Defaults to
            False.

    .. doctest:: world_test_new

        >>> from antidote import world
        >>> world.singletons.add("test", 1)
        >>> with world.test.new():
        ...     # Current world is empty
        ...     world.get[int]("test")
        Traceback (most recent call last):
        ...
        DependencyNotFoundError: 'test'
        >>> with world.test.new():
        ...     world.singletons.add("new", 2)
        >>> # Anything done within the context manager stays there.
        ... world.get[int]("new")
        Traceback (most recent call last):
        ...
        DependencyNotFoundError: 'test'

    """
    if default_providers:
        from ..._internal.utils.world import new_container
        with state.override(lambda _: new_container()):
            yield
    else:
        with state.override(lambda c: c.clone(keep_singletons=False,
                                              clone_providers=False)):
            yield


###############################
# Utilities to test providers #
###############################

@API.public
@contextmanager
def empty():
    """
    Only used to test providers.

    Creates an empty container. No initial provider are set up. This is useful to test one
    or several providers in isolation. If you're not testing provider use :py:func:`.new`
    instead.
    """
    with state.override(lambda _: RawContainer()):
        yield


@API.public
def maybe_provide_from(provider: RawProvider,
                       dependency: Hashable) -> Optional[DependencyInstance]:
    """
    Only used to test providers.

    Utility function to test _providers that have not been registered in :py:mod:`.world`.
    The current :py:class:~.core.container.DependencyContainer` will be given as if it
    were registered. This allows to test exactly what has been returned by the provider.

    Args:
        provider: provider from which the dependency should be retrieved.
        dependency: dependency to retrieve.

    Returns:
        dependency instance as returned by the provider.
    """
    if provider.is_registered:
        raise RuntimeError("Method only intended to test provider that have not "
                           "been registered in world.")
    return provider.maybe_provide(dependency, state.get_container())


__sentinel = object()


@overload
def singleton(dependency: Hashable, value) -> None: ...  # noqa: E704


@overload
def singleton(dependency: dict) -> None: ...  # noqa: E704


@API.public
def singleton(dependency: Union[dict, Hashable], value=__sentinel) -> None:
    """
    Override one or multiple dependencies with one/multiple singletons.

    .. doctest:: world_test_override_singleton

        >>> from antidote import world
        >>> world.singletons.add("test", 1)
        >>> world.get[int]("test")
        1
        >>> with world.test.clone(overridable=True):
        ...     world.test.override.singleton("test", 2)
        ...     world.test.override.singleton({'test': 2})
        ...     world.get[int]("test")
        2
        >>> # Overrides works only within the test world.
        ... world.get[int]("test")
        1

    Args:
        dependency: Singleton to declare, must be hashable. If a dict is provided, it'll
            be treated as a dictionary of singletons to add.
        value: Associated value for the dependency.

    """
    if value is __sentinel:
        if isinstance(dependency, dict):
            state.get_overridable_container().override_singletons(dependency)
        else:
            raise TypeError("If only a single argument is provided, "
                            "it must be a dictionary of singletons.")
    else:
        state.get_overridable_container().override_singletons({dependency: value})


F = TypeVar('F', bound=Callable[..., Any])


@API.public
def factory(dependency: Hashable, *, singleton: bool = False) -> Callable[[F], F]:
    """
    Override a dependency with the result of a factory. To be used as a function
    decorator. The result of the underlying function, the factory, will be used as the
    associated value of the dependency.

    .. doctest:: world_test_override_factory

        >>> from antidote import world
        >>> world.singletons.add("test", 1)
        >>> world.get[int]("test")
        1
        >>> with world.test.clone(overridable=True):
        ...     @world.test.override.factory('test', singleton=True)
        ...     def test():
        ...         return 2
        ...     world.get[int]("test")
        2
        >>> # Overrides works only within the test world.
        ... world.get[int]("test")
        1

    Args:
        dependency: Dependency to override.
        singleton: Whether the output returned by the factory should be treated as a
            singleton or not. If so, the factory will only be called once.

    .. note::

        As all override work as a single layer on top of the usual dependencies, if the
        dependency is already overridden as a singleton, the factory won't be called.
        Overrides done with a provider also have a higher priority. See
        :py:mod:`.world.test.override`.

    Returns:
        The decorated function, unchanged.
    """
    def decorate(f: 'Callable[..., Any]'):
        state.get_overridable_container().override_factory(dependency,
                                                           factory=f,
                                                           singleton=singleton)
        return f

    return decorate


P = TypeVar('P', bound=Callable[[Any], Optional[DependencyInstance]])


@API.public
def provider(p: P) -> P:
    """
    Function decorator used to declare a simplified provider to override some
    dependencies. The function will be called for each dependency and should
    return :py:class:`.core.DependencyInstance` if it can be provided or :py:obj:`None`
    if not.

    .. doctest:: world_test_override_provider

        >>> from antidote import world
        >>> from antidote.core import DependencyInstance
        >>> world.singletons.add("test", 1)
        >>> world.get[int]("test")
        1
        >>> with world.test.clone(overridable=True):
        ...     @world.test.override.provider
        ...     def test(dependency):
        ...         if dependency == 'test':
        ...             return DependencyInstance(2, singleton=True)
        ...     world.get[int]("test")
        2
        >>> # Overrides works only within the test world.
        ... world.get[int]("test")
        1

    .. note::

        As all override work as a single layer on top of the usual dependencies, singleton
        overrides won't pass through the provider. See :py:mod:`.world.test.override`.

    .. warning::

        Currently, provider overrides won't show in :py:func:`.world.debug`.

    Args:
        p: provider function.

    Returns:
        the decorated function, unchanged.
    """
    state.get_overridable_container().override_provider(p)
    return p



from .factory import FactoryProvider
from .indirect import IndirectProvider
from .lazy import Lazy, LazyProvider
from .service import ServiceProvider
from .tag import DuplicateTagError, Tag, Tagged, TagProvider



from antidote import Tag
from antidote.exceptions import (DependencyCycleError, DependencyNotFoundError,
                                 DuplicateDependencyError, DuplicateTagError,
                                 FrozenWorldError)


class Service:
    pass


def test_dependency_cycle_error():
    error = DependencyCycleError([Service, 'test', Service])

    service_info = f"{Service.__module__}.{Service.__name__}"

    for f in [str, repr]:
        assert service_info in f(error)
        assert "'test'" in f(error)


def test_dependency_not_found():
    o = object()
    error = DependencyNotFoundError(o)

    for f in [str, repr]:
        assert repr(o) in f(error)


def test_duplicate_dependency_error():
    dependency = object()
    existing_dependency = object()
    error = DuplicateDependencyError(dependency, existing_dependency)

    for f in [str, repr]:
        assert f(dependency) in f(error)
        assert f(existing_dependency) in f(error)

    message = "hello"
    error = DuplicateDependencyError(message)
    assert message in str(error)


def test_duplicate_tag_error():
    tag = Tag()
    dependency = object()

    error = DuplicateTagError(dependency, tag)
    assert str(tag) in str(error)
    assert str(dependency) in str(error)


def test_frozen_world():
    message = "my message"
    error = FrozenWorldError(message)
    assert message in str(error)
    assert message in repr(error)


def test_duplicate_dependency():
    x = object()
    y = object()
    error = DuplicateDependencyError(x, y)
    for f in [str, repr]:
        assert f(x) in f(error)
        assert f(y) in f(error)

    message = "test"
    error = DuplicateDependencyError(message)
    assert str(message) in str(error)
    assert str(message) in repr(error)



import threading
from collections import deque
from contextlib import contextmanager
from typing import (Any, Callable, cast, Deque, Dict, Hashable, List, Mapping, Optional,
                    Sequence, Tuple, Type)
from weakref import ref, ReferenceType

from .exceptions import (DependencyCycleError, DependencyInstantiationError,
                         DependencyNotFoundError, DuplicateDependencyError,
                         FrozenWorldError)
from .utils import DependencyDebug
from .._compatibility.typing import final
from .._internal import API
from .._internal.stack import DependencyStack
# PRIVATE
from .._internal.utils import FinalImmutable

_CONTAINER_REF_ATTR = "_antidote__container_ref"


@API.public
@final
class DependencyInstance(FinalImmutable):
    """
    Simple wrapper of a dependency instance given by a
    :py:class:`~.provider.Provider`.
    """
    __slots__ = ('value', 'singleton')
    value: Any
    singleton: bool

    def __init__(self, value: Any, *, singleton: bool = False):
        super().__init__(value, singleton)

    def __eq__(self, other):
        return isinstance(other, DependencyInstance) \
               and self.singleton == other.singleton \
               and self.value == other.value  # noqa: E126


@API.public
class Container:
    """
    Public interface of the container used by Antidote to handles all
    dependencies. Used in a :py:class:`~.provider.Provider` to access other
    dependencies.
    """

    def get(self, dependency: Hashable) -> Any:
        """
        Retrieves given dependency or raises a
        :py:exc:`~..exceptions.DependencyNotFoundError`.

        Args:
            dependency: Dependency to be retrieved.

        Returns:
            The dependency instance.
        """
        raise NotImplementedError()  # pragma: no cover

    def provide(self, dependency: Hashable) -> DependencyInstance:
        """
        Similar to :py:meth:`~.get` it will retrieve a dependency. However it will
        be wrapped in a :py:class:`~.DependencyInstance` if found. If not
        :py:obj:`None` will be returned. This allows to get additional information such
        as whether the dependency is a singleton or not.

        Args:
            dependency: Dependency to be retrieved.

        Returns:
            wrapped dependency instance.

        """
        raise NotImplementedError()  # pragma: no cover


###########
# PRIVATE #
###########

@API.private
class RawProvider:
    """
    Abstract base class for a Provider.

    Prefer using :py:class:`~.core.provider.Provider`
    or :py:class:`~.core.provider.StatelessProvider` which are safer
    to implement.

    :meta private:
    """

    def __init__(self):
        setattr(self, _CONTAINER_REF_ATTR, None)

    def clone(self, keep_singletons_cache: bool) -> 'RawProvider':
        raise NotImplementedError()  # pragma: no cover

    def exists(self, dependency: Hashable) -> bool:
        raise NotImplementedError()  # pragma: no cover

    def maybe_provide(self, dependency: Hashable, container: Container
                      ) -> Optional[DependencyInstance]:
        raise NotImplementedError()  # pragma: no cover

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        raise NotImplementedError()  # pragma: no cover

    @API.private
    @final
    @contextmanager
    def _ensure_not_frozen(self):
        container_ref: ReferenceType[RawContainer] = getattr(self,
                                                             _CONTAINER_REF_ATTR)
        if container_ref is None:
            yield
        else:
            container = container_ref()
            assert container is not None, "Associated container does not exist anymore."
            with container.ensure_not_frozen():
                yield

    @API.private
    @final
    def _raise_if_exists(self, dependency):
        container_ref: ReferenceType[RawContainer] = getattr(self,
                                                             _CONTAINER_REF_ATTR)
        if container_ref is not None:
            container = container_ref()
            assert container is not None, "Associated container does not exist anymore."
            container.raise_if_exists(dependency)
        else:
            if self.exists(dependency):
                raise DuplicateDependencyError(
                    f"{dependency} has already been registered in {type(self)}")

    @final
    @property
    def is_registered(self):
        return getattr(self, _CONTAINER_REF_ATTR) is not None


@API.private  # Not meant for direct use. You should go through world to manipulate it.
class RawContainer(Container):

    def __init__(self):
        self._dependency_stack = DependencyStack()
        self._singleton_lock = threading.RLock()
        self._providers: List[RawProvider] = list()
        self._singletons: Dict[Any, Any] = dict()
        self.__freeze_lock = threading.RLock()
        self.__frozen = False

    def __repr__(self):
        return f"{type(self).__name__}(providers={', '.join(map(str, self._providers))})"

    @property
    def providers(self) -> Sequence[RawProvider]:
        return self._providers.copy()

    @contextmanager
    def ensure_not_frozen(self):
        with self.__freeze_lock:
            if self.__frozen:
                raise FrozenWorldError()
            yield

    def freeze(self):
        with self.__freeze_lock:
            self.__frozen = True

    def add_provider(self, provider_cls: Type[RawProvider]):
        if not isinstance(provider_cls, type) \
                or not issubclass(provider_cls, RawProvider):
            raise TypeError(
                f"provider must be a Provider, not a {provider_cls}")

        with self.ensure_not_frozen(), self._singleton_lock:
            if any(provider_cls == type(p) for p in self._providers):
                raise ValueError(f"Provider {provider_cls} already exists")

            provider = provider_cls()
            setattr(provider, _CONTAINER_REF_ATTR, ref(self))
            self._providers.append(provider)
            self._singletons[provider_cls] = provider

    def add_singletons(self, dependencies: Mapping):
        with self.ensure_not_frozen(), self._singleton_lock:
            for k, v in dependencies.items():
                self.raise_if_exists(k)
            self._singletons.update(dependencies)

    def raise_if_exists(self, dependency: Hashable):
        with self.__freeze_lock:
            if dependency in self._singletons:
                raise DuplicateDependencyError(
                    f"{dependency!r} has already been defined as a singleton pointing "
                    f"to {self._singletons[dependency]}")

            for provider in self._providers:
                if provider.exists(dependency):
                    debug = provider.maybe_debug(dependency)
                    message = f"{dependency!r} has already been declared " \
                              f"in {type(provider)}"
                    if debug is None:
                        raise DuplicateDependencyError(message)
                    else:
                        raise DuplicateDependencyError(message + f"\n{debug.info}")

    def clone(self,
              *,
              keep_singletons: bool = False,
              clone_providers: bool = True) -> 'RawContainer':
        container = type(self)()
        with self._singleton_lock:
            if keep_singletons:
                container._singletons = self._singletons.copy()
            if clone_providers:
                for p in self._providers:
                    clone = p.clone(keep_singletons_cache=keep_singletons)
                    if clone is p \
                        or getattr(clone, _CONTAINER_REF_ATTR,
                                   None) is not None:
                        raise RuntimeError(
                            "A Provider should always return a fresh "
                            "instance when copy() is called.")

                    setattr(clone, _CONTAINER_REF_ATTR, ref(container))
                    container._providers.append(clone)
                    container._singletons[type(p)] = clone
            else:
                for p in self._providers:
                    container.add_provider(type(p))

        return container

    def debug(self, dependency: Hashable) -> DependencyDebug:
        from .._internal.utils.debug import debug_repr

        with self.__freeze_lock:
            for p in self._providers:
                debug = p.maybe_debug(dependency)
                if debug is not None:
                    return debug
            try:
                value = self._singletons[dependency]
                return DependencyDebug(f"Singleton {debug_repr(dependency)} "
                                       f"-> {value!r}",
                                       singleton=True)
            except KeyError:
                raise DependencyNotFoundError(dependency)

    def provide(self, dependency: Hashable) -> DependencyInstance:
        try:
            return DependencyInstance(self._singletons[dependency], singleton=True)
        except KeyError:
            pass
        return self._safe_provide(dependency)

    def get(self, dependency: Hashable) -> Any:
        try:
            return self._singletons[dependency]
        except KeyError:
            pass
        return self._safe_provide(dependency).value

    def _safe_provide(self, dependency: Hashable) -> DependencyInstance:
        try:
            with self._singleton_lock, \
                 self._dependency_stack.instantiating(dependency):
                try:
                    return DependencyInstance(self._singletons[dependency],
                                              singleton=True)
                except KeyError:
                    pass

                for provider in self._providers:
                    dependency_instance = provider.maybe_provide(dependency, self)
                    if dependency_instance is not None:
                        if dependency_instance.singleton:
                            self._singletons[dependency] = dependency_instance.value

                        return dependency_instance

        except DependencyCycleError:
            raise

        except Exception as e:
            raise DependencyInstantiationError(dependency) from e

        raise DependencyNotFoundError(dependency)


class OverridableRawContainer(RawContainer):
    def __init__(self):
        super().__init__()
        self.__override_lock = threading.RLock()
        # Used to differentiate singletons from the overrides and the "normal" ones.
        self.__singletons_override = dict()
        self.__factory_overrides: Dict[Any, Tuple[Callable[[], Any], bool]] = {}
        self.__provider_overrides: Deque[
            Callable[[Any], Optional[DependencyInstance]]] = deque()

    @classmethod
    def build(cls,
              original: RawContainer,
              keep_singletons: bool) -> 'OverridableRawContainer':
        container = cls()
        clone = original.clone(keep_singletons=keep_singletons)
        container._singletons = clone._singletons
        container._providers = clone._providers
        if isinstance(clone, OverridableRawContainer):
            container.__singletons_override = clone.__singletons_override
            container.__factory_overrides = clone.__factory_overrides
            container.__provider_overrides = clone.__provider_overrides
        return container

    def override_singletons(self, singletons: dict):
        if not isinstance(singletons, dict):
            raise TypeError(f"singletons must be a dict, not a {type(singletons)}")
        with self.__override_lock:
            self.__singletons_override.update(singletons)

    def override_factory(self,
                         dependency: Hashable,
                         *,
                         factory: Callable[[], Any],
                         singleton: bool):
        if not callable(factory):
            raise TypeError(f"factory must be a callable, not a {type(factory)}")
        if not isinstance(singleton, bool):
            raise TypeError(f"singleton must be a boolean, not a {type(singleton)}")
        with self.__override_lock:
            self.__factory_overrides[dependency] = (factory, singleton)

    def override_provider(self,
                          provider: Callable[[Any], Optional[DependencyInstance]]):
        if not callable(provider):
            raise TypeError(f"provider must be a callable, not a {type(provider)}")
        with self.__override_lock:
            self.__provider_overrides.appendleft(provider)  # latest provider wins

    def provide(self, dependency: Hashable) -> DependencyInstance:
        return self._safe_provide(dependency)

    def get(self, dependency: Hashable) -> Any:
        return self._safe_provide(dependency).value

    def clone(self,
              *,
              keep_singletons: bool = False,
              clone_providers: bool = True) -> 'OverridableRawContainer':
        with self._singleton_lock, self.__override_lock:
            container = cast(OverridableRawContainer,
                             super().clone(keep_singletons=keep_singletons,
                                           clone_providers=clone_providers))
            if keep_singletons:
                container.__singletons_override = self.__singletons_override.copy()
            container.__factory_overrides = self.__factory_overrides.copy()
            container.__provider_overrides = self.__provider_overrides.copy()

        return container

    def debug(self, dependency: Hashable) -> DependencyDebug:
        from .._internal.utils.debug import debug_repr

        with self.__override_lock:
            try:
                value = self.__singletons_override[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Singleton: {debug_repr(dependency)} "
                                       f"-> {value!r}",
                                       singleton=True)
            try:
                (factory, singleton) = self.__factory_overrides[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Factory: {debug_repr(dependency)} "
                                       f"-> {debug_repr(factory)}",
                                       singleton=singleton)

        return super().debug(dependency)

    def _safe_provide(self, dependency: Hashable) -> DependencyInstance:
        with self._singleton_lock, self.__override_lock:
            with self._dependency_stack.instantiating(dependency):
                try:
                    return DependencyInstance(self.__singletons_override[dependency],
                                              singleton=True)
                except KeyError:
                    pass

                try:
                    for provider in self.__provider_overrides:
                        dependency_instance = provider(dependency)
                        if dependency_instance is not None:
                            if dependency_instance.singleton:
                                self.__singletons_override[dependency] \
                                    = dependency_instance.value
                            return dependency_instance

                    try:
                        (factory, singleton) = self.__factory_overrides[dependency]
                    except KeyError:
                        pass
                    else:
                        value = factory()
                        if singleton:
                            self.__singletons_override[dependency] = value
                        return DependencyInstance(value, singleton=singleton)

                except DependencyCycleError:
                    raise

                except Exception as e:
                    raise DependencyInstantiationError(dependency) from e

            return super()._safe_provide(dependency)



import functools
import inspect
from typing import Any, Callable, Dict, get_type_hints

from ._internal import API
from ._internal.utils import AbstractMeta, FinalImmutable
from ._providers import FactoryProvider, TagProvider
from ._providers.factory import FactoryDependency
from ._providers.service import Build
from .core import Dependency, inject
from .lazy import LazyCall
from .service import service

_ABSTRACT_FLAG = '__antidote_abstract'


@API.private
class FactoryMeta(AbstractMeta):
    __factory_dependency: FactoryDependency

    def __new__(mcls, name, bases, namespace, **kwargs):
        abstract = kwargs.get('abstract')

        if '__call__' not in namespace and not abstract:
            raise TypeError(f"The class {name} must implement __call__()")

        cls = super().__new__(mcls, name, bases, namespace, **kwargs)
        if not abstract:
            cls.__factory_dependency = _configure_factory(cls)

        return cls

    @API.public
    def __rmatmul__(cls, dependency) -> Any:
        assert cls.__factory_dependency is not None
        if dependency != cls.__factory_dependency.dependency:
            raise ValueError(f"Unsupported dependency {dependency}")
        return cls.__factory_dependency

    @API.public
    def with_kwargs(cls, **kwargs) -> 'PreBuild':
        """
        Creates a new dependency based on the factory which will have the keyword
        arguments provided. If the factory provides a singleton and identical kwargs are
        used, the same instance will be returned by Antidote.

        Args:
            **kwargs: Arguments passed on to the factory.

        Returns:
            Dependency to be retrieved from Antidote.
        """
        assert cls.__factory_dependency is not None
        return PreBuild(cls.__factory_dependency, kwargs)


@API.private
@inject
def _configure_factory(cls,
                       factory_provider: FactoryProvider,
                       tag_provider: TagProvider = None):
    from .factory import Factory

    conf = getattr(cls, '__antidote__', None)
    if not isinstance(conf, Factory.Conf):
        raise TypeError(f"Factory configuration (__antidote__) is expected to be "
                        f"a {Factory.Conf}, not a {type(conf)}")

    dependency = get_type_hints(cls.__call__).get('return')
    if dependency is None:
        raise ValueError("The return annotation is necessary on __call__."
                         "It is used a the dependency.")
    if not inspect.isclass(dependency):
        raise TypeError(f"The return annotation is expected to be a class, "
                        f"not {type(dependency)}.")

    if conf.wiring is not None:
        conf.wiring.wire(cls)

    factory_id = factory_provider.register(
        dependency=dependency,
        singleton=conf.singleton,
        factory=Dependency(service(cls, singleton=True)
                           if conf.public else
                           LazyCall(cls, singleton=True))
    )

    if conf.tags is not None:
        if tag_provider is None:
            raise RuntimeError("No TagProvider registered, cannot use tags.")
        tag_provider.register(dependency=factory_id, tags=conf.tags)

    return factory_id


@API.private
class LambdaFactory:
    def __init__(self, factory: Callable, dependency_factory: FactoryDependency):
        self.__factory = factory
        self.__dependency_factory = dependency_factory
        functools.wraps(factory, updated=())(self)

    def __call__(self, *args, **kwargs):
        return self.__factory(*args, **kwargs)

    def __rmatmul__(self, dependency) -> Any:
        if dependency != self.__dependency_factory.dependency:
            raise ValueError(f"Unsupported dependency {dependency}")
        return self.__dependency_factory

    def with_kwargs(self, **kwargs) -> 'PreBuild':
        return PreBuild(self.__dependency_factory, kwargs)


@API.private
class PreBuild(FinalImmutable):
    __slots__ = ('dependency_factory', 'kwargs')
    dependency_factory: FactoryDependency
    kwargs: Dict

    def __init__(self, dependency_factory: FactoryDependency, kwargs: Dict):
        if not kwargs:
            raise ValueError("When calling with_kwargs, "
                             "at least one argument must be provided.")
        super().__init__(dependency_factory=dependency_factory, kwargs=kwargs)

    def __rmatmul__(self, dependency) -> Any:
        if dependency != self.dependency_factory.dependency:
            raise ValueError(f"Unsupported dependency {dependency}")
        return Build(self.dependency_factory, self.kwargs)



from ._internal import API
from ._providers import IndirectProvider
from ._service import ServiceMeta
from .core import inject

_ABSTRACT_FLAG = '__antidote_abstract'


@API.private
class ImplementationMeta(ServiceMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        interface = None
        if any(isinstance(b, ImplementationMeta) for b in bases):
            if kwargs.get("abstract"):
                raise ValueError("Implementation does not support abstract sub classes")

            if len(bases) < 2:
                raise TypeError("The interface to be implemented must be specified "
                                "as first base class.")
            interface, impl_base, *_ = bases

            if not isinstance(impl_base, ImplementationMeta) \
                    or sum(1 for b in bases if isinstance(b, ImplementationMeta)) > 1:
                raise TypeError("The second base class, and only this one, "
                                "must be Implementation.")

        cls = super().__new__(mcls, name, bases, namespace, **kwargs)

        if interface is not None:
            _configure_implementation(interface, cls)

        return cls


@API.private
@inject
def _configure_implementation(interface, cls, indirect_provider: IndirectProvider):
    indirect_provider.register_static(interface, cls)



from ._internal import API


@API.public
def is_compiled() -> bool:
    """
    Whether current Antidote implementations is the compiled (Cython) version or not
    """
    from ._internal.wrapper import compiled
    return compiled



import builtins
import collections.abc as c_abc
import inspect
from typing import (Any, Dict, Iterable, Mapping, Set, TYPE_CHECKING, Union)

from .._internal import API
from .._internal.argspec import Arguments
from .._internal.wrapper import (build_wrapper, Injection, InjectionBlueprint, is_wrapper)

_BUILTINS_TYPES = {e for e in builtins.__dict__.values() if isinstance(e, type)}

if TYPE_CHECKING:
    from .injection import DEPENDENCIES_TYPE


@API.private  # Use inject instead
def raw_inject(func=None,
               *,
               arguments: Arguments = None,
               dependencies: 'DEPENDENCIES_TYPE' = None,
               use_names: Union[bool, Iterable[str]] = None,
               use_type_hints: Union[bool, Iterable[str]] = None):
    def _inject(wrapped):
        nonlocal arguments

        if inspect.isclass(wrapped):
            # @inject on a class would not return a class which is
            # counter-intuitive.
            raise TypeError("Classes cannot be wrapped with @inject. "
                            "Consider using @wire")
        if not (callable(wrapped) or isinstance(wrapped, (classmethod, staticmethod))):
            raise TypeError(f"wrapped object {wrapped} is not callable "
                            f"neither a class/static method")

        # if the function has already its dependencies injected, no need to do
        # it twice.
        if is_wrapper(wrapped):
            return wrapped

        if arguments is None:
            arguments = Arguments.from_callable(wrapped)

        blueprint = _build_injection_blueprint(
            arguments=arguments,
            dependencies=dependencies,
            use_names=use_names,
            use_type_hints=use_type_hints
        )

        # If nothing can be injected, just return the existing function without
        # any overhead.
        if blueprint.is_empty():
            return wrapped

        return build_wrapper(blueprint=blueprint, wrapped=wrapped)

    return func and _inject(func) or _inject


@API.private
def _build_injection_blueprint(arguments: Arguments,
                               dependencies: 'DEPENDENCIES_TYPE' = None,
                               use_names: Union[bool, Iterable[str]] = None,
                               use_type_hints: Union[bool, Iterable[str]] = None,
                               ) -> InjectionBlueprint:
    """
    Construct a InjectionBlueprint with all the necessary information about
    the arguments for dependency injection. Storing it avoids significant
    execution overhead./

    Used by inject()
    """
    use_names = use_names if use_names is not None else False
    use_type_hints = use_type_hints if use_type_hints is not None else True

    arg_to_dependency = _build_arg_to_dependency(arguments, dependencies)
    type_hints = _build_type_hints(arguments, use_type_hints)
    dependency_names = _build_dependency_names(arguments, use_names)

    resolved_dependencies = [
        arg_to_dependency.get(
            arg.name,
            type_hints.get(arg.name,
                           arg.name if arg.name in dependency_names else None)
        )
        for arg in arguments
    ]

    return InjectionBlueprint(tuple(
        Injection(arg_name=arg.name,
                  required=not arg.has_default,
                  dependency=dependency)
        for arg, dependency in zip(arguments, resolved_dependencies)
    ))


@API.private
def _build_arg_to_dependency(arguments: Arguments,
                             dependencies: 'DEPENDENCIES_TYPE' = None
                             ) -> Dict[str, Any]:
    from .injection import Arg
    if dependencies is None:
        arg_to_dependency: Mapping = {}
    elif isinstance(dependencies, str):
        if "{arg_name}" not in dependencies:
            raise ValueError("Missing formatting parameter {arg_name} in dependencies. "
                             "If you really want a constant injection, "
                             "consider using a defaultdict.")
        arg_to_dependency = {arg.name: dependencies.format(arg_name=arg.name)
                             for arg in arguments.without_self}
    elif callable(dependencies):
        arg_to_dependency = {arg.name: dependencies(Arg(arg.name, arg.type_hint))
                             for arg in arguments.without_self}
    elif isinstance(dependencies, c_abc.Mapping):
        _check_valid_arg_names(dependencies.keys(), arguments)
        arg_to_dependency = dependencies
    elif isinstance(dependencies, c_abc.Iterable):
        # convert to Tuple in case we cannot iterate more than once.
        dependencies = tuple(dependencies)
        if len(arguments.without_self) < len(dependencies):
            raise ValueError(f"More dependencies ({dependencies}) were provided "
                             f"than arguments ({arguments})")
        arg_to_dependency = {arg.name: dependency
                             for arg, dependency
                             in zip(arguments.without_self, dependencies)}
    else:
        raise TypeError(f'Only a mapping or a iterable is supported for '
                        f'dependencies, not {type(dependencies)!r}')

    # Remove any None as they would hide type_hints and use_names.
    return {
        k: v
        for k, v in arg_to_dependency.items()
        if v is not None
    }


@API.private
def _build_type_hints(arguments: Arguments,
                      use_type_hints: Union[bool, Iterable[str]]) -> Dict[str, Any]:
    if use_type_hints is True:
        type_hints = {arg.name: arg.type_hint for arg in arguments.without_self}
    elif use_type_hints is False:
        return {}
    elif isinstance(use_type_hints, c_abc.Iterable):
        # convert to Tuple in case we cannot iterate more than once.
        use_type_hints = tuple(use_type_hints)
        _check_valid_arg_names(use_type_hints, arguments)

        type_hints = {name: arguments[name].type_hint for name in use_type_hints}

    else:
        raise TypeError(f"Only an iterable or a boolean is supported for "
                        f"use_type_hints, not {type(use_type_hints)!r}")

    for arg_name in list(type_hints.keys()):
        type_hint = type_hints[arg_name]
        if getattr(type_hint, '__origin__', None) is Union \
                and len(type_hint.__args__) == 2:
            a = type_hint.__args__
            if isinstance(None, a[1]):
                type_hints[arg_name] = a[0]

    # Any object from builtins or typing do not carry any useful information
    # and thus must not be used as dependency IDs. So they might as well be
    # skipped entirely. Moreover they hide use_names.
    return {
        arg_name: type_hint
        for arg_name, type_hint in type_hints.items()
        if getattr(type_hint, '__module__', '') != 'typing'
           and type_hint not in _BUILTINS_TYPES  # noqa
           and type_hint is not None  # noqa
    }


@API.private
def _build_dependency_names(arguments: Arguments,
                            use_names: Union[bool, Iterable[str]]) -> Set[str]:
    if use_names is False:
        return set()
    elif use_names is True:
        return {arg.name for arg in arguments.without_self}
    elif isinstance(use_names, c_abc.Iterable):
        # convert to Tuple in case we cannot iterate more than once.
        use_names = tuple(use_names)
        _check_valid_arg_names(use_names, arguments)
        return set(use_names)
    else:
        raise TypeError(f'Only an iterable or a boolean is supported for '
                        f'use_names, not {type(use_names)!r}')


@API.private
def _check_valid_arg_names(names: Iterable[str], arguments: Arguments):
    for name in names:
        if not isinstance(name, str):
            raise TypeError(f"Expected argument name (str), "
                            f"not {type(name)}")
        if name not in arguments:
            raise ValueError(f"Unknown argument '{name}'")
        if arguments.has_self and name == arguments[0].name:
            raise ValueError(f"Cannot inject first argument "
                             f"('{arguments[0]}') of a method")



import weakref
from typing import Callable, TYPE_CHECKING, Union

from ._compatibility.typing import final
from ._internal import API
from ._internal.utils import debug_repr, FinalImmutable, short_id
from ._providers import Lazy
from .core import Container, DependencyInstance
from .core.utils import DependencyDebug

if TYPE_CHECKING:
    from .lazy import LazyMethodCall


@API.private
@final
class LazyCallWithArgsKwargs(FinalImmutable, Lazy):
    """
    :meta private:
    """
    __slots__ = ('func', 'singleton', 'args', 'kwargs')
    func: Callable
    singleton: bool
    args: tuple
    kwargs: dict

    def __antidote_debug_repr__(self):
        s = f"Lazy: {debug_repr(self.func)}(*{self.args}, **{self.kwargs})"
        if self.singleton:
            s += f"  #{short_id(self)}"
        return s

    def debug_info(self) -> DependencyDebug:
        return DependencyDebug(self.__antidote_debug_repr__(),
                               singleton=self.singleton,
                               wired=[self.func])

    def lazy_get(self, container: Container) -> DependencyInstance:
        return DependencyInstance(self.func(*self.args, **self.kwargs),
                                  singleton=self.singleton)


@API.private
@final
class LazyMethodCallWithArgsKwargs(FinalImmutable):
    """
    :meta private:
    """
    __slots__ = ('method_name', 'singleton', 'args', 'kwargs', '_cache_attr')
    method_name: str
    singleton: bool
    args: tuple
    kwargs: dict
    _cache_attr: str

    def __init__(self, method_name: str, singleton: bool, args: tuple, kwargs: dict):
        super().__init__(method_name, singleton, args, kwargs,
                         f"__antidote_dependency_{hex(id(self))}")

    def __get__(self, instance, owner):
        if instance is None:
            try:
                return getattr(owner, self._cache_attr)
            except AttributeError:
                dependency = LazyMethodCallDependency(self,
                                                      weakref.ref(owner),
                                                      self.singleton)
                setattr(owner, self._cache_attr, dependency)
                return dependency
        return getattr(instance, self.method_name)(*self.args, **self.kwargs)

    def __str__(self):
        s = f"Lazy Method: {self.method_name}(*{self.args}, **{self.kwargs})"
        if self.singleton:
            s += f"  #{short_id(self)}"
        return s


@API.private
@final
class LazyMethodCallDependency(FinalImmutable, Lazy):
    """
    :meta private:
    """
    __slots__ = ('descriptor', 'owner_ref', 'singleton')
    descriptor: 'Union[LazyMethodCall, LazyMethodCallWithArgsKwargs]'
    owner_ref: weakref.ReferenceType
    singleton: bool

    def debug_info(self) -> DependencyDebug:
        owner = self.owner_ref()
        assert owner is not None
        method = self.descriptor.method_name
        return DependencyDebug(str(self.descriptor),
                               singleton=self.descriptor.singleton,
                               wired=[getattr(owner, method)],
                               dependencies=[owner])

    def lazy_get(self, container: Container) -> DependencyInstance:
        owner = self.owner_ref()
        assert owner is not None
        return DependencyInstance(self.descriptor.__get__(container.get(owner), owner),
                                  singleton=self.singleton)



from antidote import world
from antidote.core import Dependency, DependencyInstance
from antidote.core.utils import DependencyDebug


def test_dependency():
    with world.test.empty():
        world.singletons.add('x', object())
        d = Dependency('x')
        assert d.value == 'x'
        assert d.get() is world.get('x')


def test_dependency_instance():
    ref = DependencyInstance("test", singleton=True)
    assert ref == DependencyInstance("test", singleton=True)
    assert ref != DependencyInstance("test2", singleton=True)
    assert ref != DependencyInstance("test", singleton=False)


def test_dependency_debug():
    ref = DependencyDebug(info="info", singleton=True, wired=[1], dependencies=[2])
    assert ref == DependencyDebug(info="info", singleton=True, wired=[1],
                                  dependencies=[2])
    assert ref != DependencyDebug(info="info2", singleton=True, wired=[1],
                                  dependencies=[2])
    assert ref != DependencyDebug(info="info", singleton=False, wired=[1],
                                  dependencies=[2])
    assert ref != DependencyDebug(info="info", singleton=True, wired=[10],
                                  dependencies=[2])
    assert ref != DependencyDebug(info="info", singleton=True, wired=[1],
                                  dependencies=[20])



import collections.abc as c_abc
from typing import Callable, Iterable, Optional, overload, Tuple, TypeVar, Union

from ._compatibility.typing import final
from ._internal import API
from ._internal.utils import Copy, FinalImmutable
from ._providers import Tag
from ._service import ServiceMeta
from .core.exceptions import DuplicateDependencyError
from .core.wiring import AutoWire, Wiring, WithWiringMixin

C = TypeVar('C', bound=type)


@API.public
class Service(metaclass=ServiceMeta, abstract=True):
    """
    Abstract base class for a service. It can be retrieved from Antidote simply with its
    class. It will only be instantiated when necessary. By default a service will be
    a singleton and :py:meth:`.__init__` will be wired (injected).

    .. doctest:: helpers_Service

        >>> from antidote import Service, world
        >>> class MyService(Service):
        ...     pass
        >>> world.get[MyService]()
        <MyService ...>

    For customization use :py:attr:`.__antidote__`:

    .. doctest:: helpers_Service_v2

        >>> from antidote import Service, world
        >>> class MyService(Service):
        ...     __antidote__ = Service.Conf(singleton=False) \\
        ...         .with_wiring(use_names=True)

    One can customize the instantiation and use the same service with different
    configuration:

    .. doctest:: helpers_Service_v3

        >>> from antidote import Service, world, inject
        >>> class MyService(Service):
        ...     def __init__(self, name = 'default'):
        ...         self.name = name
        >>> world.get[MyService]().name
        'default'
        >>> s = world.get[MyService](MyService.with_kwargs(name='perfection'))
        >>> s.name
        'perfection'
        >>> # The same instance will be returned for those keywords as MyService is
        ... # defined as a singleton.
        ... s is world.get(MyService.with_kwargs(name='perfection'))
        True
        >>> # You can also keep the dependency and re-use it
        ... PerfectionService = MyService.with_kwargs(name='perfection')
        >>> @inject(dependencies=dict(service=PerfectionService))
        ... def f(service):
        ...     return service
        >>> f() is s
        True

    """

    @final
    class Conf(FinalImmutable, WithWiringMixin):
        """
        Immutable service configuration. To change parameters on a existing instance, use
        either method :py:meth:`.copy` or
        :py:meth:`.core.wiring.WithWiringMixin.with_wiring`.
        """
        __slots__ = ('wiring', 'singleton', 'tags', 'factory')
        wiring: Optional[Wiring]
        singleton: bool
        tags: Optional[Tuple[Tag]]

        def __init__(self,
                     *,
                     wiring: Optional[Union[Wiring, AutoWire]] =
                     Wiring(methods=['__init__'],
                            ignore_missing_method=['__init__']),
                     singleton: bool = True,
                     tags: Optional[Iterable[Tag]] = None):
            """
            Args:
                wiring: Wiring to be applied on the service. By default only
                    :code:`__init__()` will be wired. Unless :py:attr:`.factory` is class
                    method name, in which case only the latter would be wired. To
                    deactivate any wiring at all use :py:obj:`None`.
                singleton: Whether the service is a singleton or not. A singleton is
                    instantiated only once. Defaults to :py:obj:`True`
                tags: Iterable of :py:class:`~.._providers.tag.Tag` tagging to the
                      service.
            """
            if not isinstance(singleton, bool):
                raise TypeError(f"singleton can be a boolean, "
                                f"but not a {type(singleton)}")
            if not (tags is None or isinstance(tags, c_abc.Iterable)):
                raise TypeError(f"tags can be None or an iterable of strings/Tags, "
                                f"but not a {type(tags)}")
            elif tags is not None:
                tags = tuple(tags)
                if not all(isinstance(t, Tag) for t in tags):
                    raise TypeError(f"Not all tags were instances of Tag: {tags}")
            if not (wiring is None or isinstance(wiring, Wiring)):
                raise TypeError(f"wiring can be None or a Wiring, "
                                f"but not a {type(wiring)}")
            super().__init__(wiring=wiring, singleton=singleton, tags=tags)

        def copy(self,
                 *,
                 wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL,
                 singleton: Union[bool, Copy] = Copy.IDENTICAL,
                 tags: Union[Optional[Iterable[Tag]], Copy] = Copy.IDENTICAL):
            """
            Copies current configuration and overrides only specified arguments.
            Accepts the same arguments as :py:meth:`.__init__`
            """
            return Service.Conf(
                wiring=self.wiring if wiring is Copy.IDENTICAL else wiring,
                singleton=self.singleton if singleton is Copy.IDENTICAL else singleton,
                tags=self.tags if tags is Copy.IDENTICAL else tags
            )

    __antidote__: Conf = Conf()
    """
    Configuration of the service. Defaults to wire :py:meth:`.__init__`.
    """


@overload
def service(klass: C,  # noqa: E704  # pragma: no cover
            *,
            singleton: bool = True,
            tags: Iterable[Tag] = None
            ) -> C: ...


@overload
def service(*,  # noqa: E704  # pragma: no cover
            singleton: bool = True,
            tags: Iterable[Tag] = None
            ) -> Callable[[C], C]: ...


@API.experimental
def service(klass=None,
            *,
            singleton: bool = True,
            tags: Iterable[Tag] = None):
    """
    Register a service: the class itself is the dependency. Prefer using
    :py:class:`.Service` when possible, it provides more features such as
    :py:meth:`.Service.with_kwargs` and wiring. This decorator is intended for
    registration of class which cannot inherit :py:class:`.Service`. No changes will
    on the class.

    .. doctest:: helpers_register

        >>> from antidote import service
        >>> @service
        ... class CannotInheritService:
        ...     pass

    .. note::

        If your wish to declare to register an external class to Antidote, prefer using
        a factory with either :py:class:`~.extension.factory.Factory` or
        :py:func:`~.extension.factory.factory`.

    Args:
        klass: Class to register as a dependency. It will be instantiated  only when
            requested.
        singleton: If True, the service will be a singleton and instantiated only once.
        tags: Iterable of :py:class:`~.._providers.tag.Tag` applied to the service.

    Returns:
        The class or the class decorator.

    """

    def reg(cls):
        from ._service import _configure_service

        if issubclass(cls, Service):
            raise DuplicateDependencyError(f"{cls} is already defined as a dependency "
                                           f"by inheriting {Service}")

        _configure_service(cls,
                           conf=Service.Conf(wiring=None,
                                             singleton=singleton,
                                             tags=tags))

        return cls

    return klass and reg(klass) or reg



import pytest

from antidote import const, Constants, Wiring, world
from antidote._providers import LazyProvider, ServiceProvider
from antidote.exceptions import DependencyNotFoundError


class A:
    pass


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(LazyProvider)
        world.provider(ServiceProvider)
        yield


def test_lazy_constants():
    class Config(Constants):
        A = 'a'
        B = 'b'

        def get(self, key):
            return key * 2

    assert world.get(Config.A) == 'aa'
    assert world.get(Config.B) == 'bb'

    conf = Config()
    assert conf.A == 'aa'
    assert conf.B == 'bb'


@pytest.mark.parametrize('auto_cast, a, b, c', [
    pytest.param(True, 109, 3.14, '199', id='True'),
    pytest.param(False, '109', '3.14', 199, id='False'),
    pytest.param([str, int], 109, '3.14', '199', id='(str, int)'),
    pytest.param([float], '109', 3.14, 199, id='(float,)'),
])
def test_auto_cast(auto_cast, a, b, c):
    D = object()

    class Config(Constants):
        __antidote__ = Constants.Conf(auto_cast=auto_cast)

        A = const[int]('a')
        B = const[float]('b')
        C = const[str]('c')
        D = const[dict]('d')

        def get(self, key):
            if key == 'a':
                return '109'
            if key == 'b':
                return '3.14'
            if key == 'c':
                return 199
            if key == 'd':
                return D

    assert world.get(Config.A) == a
    assert world.get(Config.B) == b
    assert world.get(Config.C) == c
    assert world.get(Config.D) is D

    assert Config().A == a
    assert Config().B == b
    assert Config().C == c
    assert Config().D is D


def test_no_rule():
    class Config(Constants):
        __antidote__ = Constants.Conf(is_const=None)
        A = 'a'

        def get(self, key):
            return key * 2

    with pytest.raises(DependencyNotFoundError):
        world.get(Config.A)

    conf = Config()
    assert conf.A == 'a'


def test_custom_rule():
    class Config(Constants):
        __antidote__ = Constants.Conf(is_const=lambda name: name.startswith('hello'))
        A = 'a'
        helloA = 'a'

        def get(self, key):
            return key * 2

    assert world.get(Config.helloA) == 'aa'
    with pytest.raises(DependencyNotFoundError):
        world.get(Config.A)

    conf = Config()
    assert conf.A == 'a'
    assert conf.helloA == 'aa'


def test_const():
    class Config(Constants):
        __antidote__ = Constants.Conf(is_const=None)

        a = const('1')
        b = const[int]('2')

        def get(self, key):
            return int(key)

    assert world.get(Config.a) == 1
    assert world.get(Config.b) == 2

    conf = Config()
    assert conf.a == 1
    assert conf.b == 2


def test_const_with_is_const():
    class Config(Constants):
        A = const('1')
        B = const[int]('2')

        def get(self, key):
            return int(key)

    assert world.get(Config.A) == 1
    assert world.get(Config.B) == 2

    conf = Config()
    assert conf.A == 1
    assert conf.B == 2


def test_invalid_lazy_method():
    with pytest.raises(TypeError):
        class Config(Constants):
            A = 'a'


def test_private_attribute():
    class Config(Constants):
        _A = 'a'
        b = 'b'

        def get(self, key):
            return key * 2

    conf = Config()
    assert 'a' == conf._A
    assert 'b' == conf.b


def test_public():
    with world.test.clone():
        class Config(Constants):
            def get(self, key):
                return key

        with pytest.raises(DependencyNotFoundError):
            world.get(Config)

    with world.test.clone():
        class Config(Constants):
            __antidote__ = Constants.Conf(public=False)

            def get(self, key):
                return key

        with pytest.raises(DependencyNotFoundError):
            world.get(Config)

    with world.test.clone():
        class Config(Constants):
            __antidote__ = Constants.Conf(public=True)

            def get(self, key):
                return key

        assert isinstance(world.get(Config), Config)


def test_invalid_conf():
    with pytest.raises(TypeError, match=".*__antidote__.*"):
        class Config(Constants):
            __antidote__ = object()

            def get(self, key):
                pass


def test_no_subclass_of_service():
    class Dummy(Constants):
        def get(self, key):
            pass

    with pytest.raises(TypeError, match=".*abstract.*"):
        class SubDummy(Dummy):
            def get(self, key):
                pass


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(wiring=object()), pytest.raises(TypeError, match=".*wiring.*")),
    (dict(is_const=object()), pytest.raises(TypeError, match=".*is_const.*")),
    (dict(public=object()), pytest.raises(TypeError, match=".*public.*")),
    (dict(auto_cast=object()), pytest.raises(TypeError, match=".*auto_cast.*")),
    (dict(auto_cast=['1']), pytest.raises(TypeError, match=".*auto_cast.*")),
])
def test_conf_error(kwargs, expectation):
    with expectation:
        Constants.Conf(**kwargs)


@pytest.mark.parametrize('kwargs', [
    dict(wiring=Wiring(methods=['method'])),
    dict(public=True),
    dict(is_const=lambda name: False),
    dict(auto_cast=frozenset((str,))),
])
def test_conf_copy(kwargs):
    conf = Constants.Conf().copy(**kwargs)
    for k, v in kwargs.items():
        assert getattr(conf, k) == v



import pytest

from antidote.core.container import (DependencyInstance, OverridableRawContainer,
                                     RawContainer)
from antidote.core.exceptions import DependencyCycleError, DependencyInstantiationError
from antidote.exceptions import (DependencyNotFoundError)
from .utils import DummyProvider


@pytest.fixture
def original_container():
    c = RawContainer()
    c.add_provider(DummyProvider)
    c.get(DummyProvider).data = {'name': 'Antidote'}
    c.add_singletons({'singleton': 'yes'})
    return c


@pytest.fixture
def container(original_container):
    return OverridableRawContainer.build(original_container, keep_singletons=True)


@pytest.mark.parametrize('keep_singletons', [True, False])
def test_build(original_container: RawContainer, keep_singletons: bool):
    container = OverridableRawContainer.build(original_container,
                                              keep_singletons=keep_singletons)
    assert container.get('name') == "Antidote"
    if keep_singletons:
        assert container.get('singleton') == 'yes'
    else:
        with pytest.raises(DependencyNotFoundError):
            container.get('singleton')


def test_override_singletons(container: OverridableRawContainer):
    assert container.get('name') == "Antidote"
    assert container.get('singleton') == 'yes'

    for dep in ['name', 'singleton']:
        container.override_singletons({dep: 'different'})
        assert container.get(dep) == 'different'
        container.override_singletons({dep: 'different-v2'})
        assert container.get(dep) == 'different-v2'
        assert container.provide(dep) == DependencyInstance('different-v2',
                                                            singleton=True)


def test_override_factory(container: OverridableRawContainer):
    assert container.get('name') == "Antidote"
    assert container.get('singleton') == 'yes'

    for dep in ['name', 'singleton']:
        container.override_factory(dep, factory=lambda: 'different', singleton=False)
        assert container.get(dep) == 'different'
        container.override_factory(dep, factory=lambda: 'different-v2', singleton=False)
        assert container.get(dep) == 'different-v2'
        assert container.provide(dep) == DependencyInstance('different-v2',
                                                            singleton=False)

        container.override_factory(dep, factory=lambda: object(), singleton=True)
        value = container.get(dep)
        assert not isinstance(value, str)
        assert container.get(dep) is value

        container.override_factory(dep, factory=lambda: object(), singleton=True)
        # didn't change anything, because it's now a singleton.
        assert container.get(dep) is value

        container.override_singletons({dep: 'Hello'})
        assert container.get(dep) == 'Hello'


def test_override_provider(container: OverridableRawContainer):
    assert container.get('name') == "Antidote"
    assert container.get('singleton') == 'yes'

    for dep in ['name', 'singleton']:
        container.override_provider(
            lambda x: DependencyInstance('different') if x == dep else None)
        assert container.get(dep) == 'different'
        container.override_provider(
            lambda x: DependencyInstance('different-v2') if x == dep else None)
        assert container.get(dep) == 'different-v2'
        assert container.provide(dep) == DependencyInstance('different-v2',
                                                            singleton=False)

        with pytest.raises(DependencyNotFoundError):
            container.get('x')  # new provider does not bring any other value

        container.override_provider(
            lambda x: DependencyInstance(object(),
                                         singleton=True) if x == dep else None)
        value = container.get(dep)
        assert not isinstance(value, str)
        assert container.get(dep) is value

        container.override_provider(
            lambda x: DependencyInstance(object(),
                                         singleton=True) if x == dep else None)
        # didn't change anything, because it's now a singleton.
        assert container.get(dep) is value

        container.override_singletons({dep: 'Hello'})
        assert container.get(dep) == 'Hello'


def test_invalid_override_singletons(container: OverridableRawContainer):
    with pytest.raises(TypeError, match='.*singletons.*'):
        container.override_singletons(object())


def test_invalid_override_factory(container: OverridableRawContainer):
    with pytest.raises(TypeError, match='.*factory.*'):
        container.override_factory(object(), factory=object(), singleton=True)
    with pytest.raises(TypeError, match='.*singleton.*'):
        container.override_factory(object(), factory=lambda: None, singleton=object())


def test_invalid_override_provider(container: OverridableRawContainer):
    with pytest.raises(TypeError, match='.*provider.*'):
        container.override_provider(object())


def test_errors(container: OverridableRawContainer):
    container.override_factory('cycle', factory=lambda: container.get('cycle'),
                               singleton=False)
    with pytest.raises(DependencyCycleError):
        container.get('cycle')

    def factory():
        raise RuntimeError()

    container.override_factory('error', factory=factory, singleton=False)
    with pytest.raises(DependencyInstantiationError):
        container.get('error')



from typing import Hashable, overload, Type, TypeVar, Union

from .._internal import API
from .._internal.state import init
from .._internal.utils.world import WorldGet, WorldLazy
from ..core.container import RawProvider

# Create the global container
init()

__sentinel = object()

# API.public
get = WorldGet()
get.__doc__ = """
Used to retrieve a dependency from Antidote. A type hint can be provided and the result
will cast to match it. It follows the same philosophy as mypy and will NOT enforce it
with a check.

Returns:
    Retrieves given dependency or raises a :py:exc:`~.exceptions.DependencyNotFoundError`

.. doctest:: world_get

    >>> from antidote import world, Service
    >>> world.singletons.add('dep', 1)
    >>> world.get('dep')
    1
    >>> # mypy will treat x as a int
    ... x = world.get[int]('dep')
    >>> class Dummy(Service):
    ...     pass
    >>> # To avoid repetition, if a type hint is provided but no argument, it
    ... # will retrieve the type hint itself.
    ... world.get[Dummy]()
    <Dummy ...>

"""

# API.public
lazy = WorldLazy()
lazy.__doc__ = """
Used to retrieves lazily a dependency. A type hint can be provided and the retrieved
instance will be cast to match it. It follows the same philosophy as mypy and will NOT
enforce it with a check.

Returns:
    Dependency: wrapped dependency.

.. doctest:: world_lazy

    >>> from antidote import world, Service
    >>> world.singletons.add('dep', 1)
    >>> dep = world.lazy('dep')
    >>> dep
    Dependency(value='dep')
    >>> dep.get()
    1
    >>> # mypy will treat x as a int
    ... x = world.lazy[int]('dep').get()
    >>> class Dummy(Service):
    ...     pass
    >>> # To avoid repetition, if a type hint is provided but no argument, it
    ... # will retrieve the type hint itself.
    ... world.lazy[Dummy]().get()
    <Dummy ...>

"""

P = TypeVar('P', bound=Type[RawProvider])


@API.public
def provider(p: P) -> P:
    """
    Class decorator used to register a new
    :py:class:`~.core.provider.Provider`.

    .. doctest:: world_provider

        >>> from antidote import world
        >>> from antidote.core import Provider
        >>> @world.provider
        ... class MyProvider(Provider):
        ...     '''Implementation missing'''

    Args:
        p: Provider class

    Returns:
        The same class, unmodified.

    """
    import inspect
    from .._internal import state
    from ..core.container import RawProvider

    if inspect.isclass(p) and issubclass(p, RawProvider):
        state.get_container().add_provider(p)
        return p

    raise TypeError(f"Provider must be a subclass of "
                    f"RawProvider, not {p}")


@API.public
def freeze() -> None:
    """
    Freezes Antidote. No additional dependencies can be added and singletons
    cannot be changed anymore. Registered singleton dependencies that have not yet
    been instantiated will not be impacted.

    It can be used to ensure that at runtime, once everything is initialized, nothing
    may be changed anymore.

    .. doctest:: world_freeze

        >>> from antidote import world
        >>> world.freeze()
        >>> world.singletons.add('dependency', 1)
        Traceback (most recent call last):
        ...
        FrozenWorldError

    """

    from .._internal import state
    state.get_container().freeze()


@overload
def add(dependency: Hashable, value) -> None: ...  # noqa: E704


@overload
def add(dependency: dict) -> None: ...  # noqa: E704


@API.public
def add(dependency: Union[dict, Hashable], value=__sentinel) -> None:
    """
    Declare one or multiple singleton dependencies with its associated instance.

    .. doctest:: world_singletons_add

        >>> from antidote import world
        >>> world.singletons.add("test", 1)
        >>> world.get[int]("test")
        1
        >>> world.singletons.add({'host': 'example.com', 'port': 80})
        >>> world.get[str]("host")
        'example.com'

    Args:
        dependency: Singleton to declare, must be hashable. If a dict is provided, it'll
            be treated as a dictionary of singletons to add.
        value: Associated value for the dependency.

    """
    from .._internal import state
    if value is __sentinel:
        if isinstance(dependency, dict):
            state.get_container().add_singletons(dependency)
        else:
            raise TypeError("If only a single argument is provided, "
                            "it must be a dictionary of singletons.")
    else:
        state.get_container().add_singletons({dependency: value})


@API.experimental
def debug(dependency, *, depth: int = -1) -> str:
    """
    To help you debug issues you may encounter with your injections, this will provide
    a tree representing all the dependencies that will be retrieved by Antidote when
    instantiating the specified dependency. It will also show whether those are singletons
    or if there are any cyclic dependencies.

    If the specified dependency is a wired class, the wired methods will also be
    presented.

    .. doctest:: world_debug

        >>> from antidote import world
        >>> world.singletons.add("test", 1)
        >>> print(world.debug('test'))
        Singleton 'test' -> 1
        <BLANKLINE>

    .. note::

        It will NOT instantiate any dependencies, hence with :py:func:`.implementation`
        for example it won't run the function to determine which implementation must be
        used.

    Args:
        dependency: Root of the dependency tree.
        depth: Maximum depth of the result tree. Defaults to -1 which implies not limit.

    Returns:

    """
    from .._internal.state import get_container
    from .._internal.utils.debug import tree_debug_info

    return tree_debug_info(get_container(), dependency, depth)



from typing import Any, Callable, cast, Optional, Type, TypeVar

from ._compatibility.typing import final, Protocol
from ._internal import API
from ._internal.utils import AbstractMeta, FinalImmutable, FinalMeta
from ._providers.lazy import FastLazyConst

T = TypeVar('T')

_CONST_CONSTRUCTOR_METHOD = 'get'


@API.private
@final
class MakeConst(metaclass=FinalMeta):
    def __call__(self, value: Any) -> Any:
        # Not true yet, but will be changed by ConstantsMeta
        return LazyConstToDo(value, None)

    def __getitem__(self, tpe: Type[T]) -> Callable[[Any], T]:
        def f(value: Any) -> T:
            return cast(T, LazyConstToDo(value, tpe))

        return f


@API.private
@final
class LazyConstToDo(FinalImmutable):
    __slots__ = ('value', 'type_')
    value: object
    type_: Optional[type]


@API.private
class ConstantsMeta(AbstractMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        cls = super().__new__(mcls, name, bases, namespace, **kwargs)
        if not kwargs.get('abstract'):
            _configure_constants(cls)
        return cls


@API.private
def _configure_constants(cls):
    from .constants import Constants
    from .lazy import LazyCall
    from .service import service

    conf = getattr(cls, '__antidote__', None)
    if not isinstance(conf, Constants.Conf):
        raise TypeError(f"Constants configuration (__antidote__) is expected to be a "
                        f"{Constants.Conf}, not a {type(conf)}")

    method = getattr(cls, _CONST_CONSTRUCTOR_METHOD, None)
    if method is None:
        raise TypeError(
            f"{cls} does not implement the lazy method '{_CONST_CONSTRUCTOR_METHOD}'")

    if conf.wiring is not None:
        conf.wiring.wire(cls)

    if conf.public:
        dependency = service(cls, singleton=True)
    else:
        dependency = LazyCall(cls, singleton=True)

    # TODO: Waiting for a fix: https://github.com/python/mypy/issues/6910
    is_const = cast(Callable[[str], bool], getattr(conf, 'is_const'))
    for name, v in list(cls.__dict__.items()):
        if isinstance(v, LazyConstToDo):
            setattr(cls,
                    name,
                    LazyConst(name,
                              dependency,
                              _CONST_CONSTRUCTOR_METHOD,
                              v.value,
                              v.type_ if v.type_ in conf.auto_cast else None))
        elif is_const(name):
            setattr(cls, name, LazyConst(name,
                                         dependency,
                                         _CONST_CONSTRUCTOR_METHOD,
                                         v))


@API.private
@final
class LazyConst(FinalImmutable):
    __slots__ = ('name', 'dependency', 'method_name', 'value', 'cast', '_cache')
    name: str
    dependency: object
    method_name: str
    value: object
    cast: Callable[[Any], Any]
    _cache: str

    def __init__(self, name: str, dependency, method_name: str, value,
                 cast: Callable[[Any], Any] = None):
        super().__init__(
            name=name,
            dependency=dependency,
            method_name=method_name,
            value=value,
            cast=cast or (lambda x: x),
            _cache=f"__antidote_dependency_{hex(id(self))}"
        )

    def __get__(self, instance, owner):
        if instance is None:
            try:
                return getattr(owner, self._cache)
            except AttributeError:
                # TODO: Waiting for a fix: https://github.com/python/mypy/issues/6910
                _cast = cast(Callable[[Any], Any], getattr(self, 'cast'))
                dependency = FastLazyConst(self.name,
                                           self.dependency,
                                           self.method_name,
                                           self.value,
                                           _cast)
                setattr(owner, self._cache, dependency)
                return dependency
        _cast = cast(Callable[[Any], Any], getattr(self, 'cast'))
        return _cast(getattr(instance, self.method_name)(self.value))
